#
#  BSD LICENSE
#
#  Copyright (c) Crane Chu <cranechu@gmail.com>
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#    * Neither the name of Intel Corporation nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# -*- coding: utf-8 -*-


import pytest
import nvme as d

import time
import logging


TEST_SCALE = 10    #1, 10


# trim
@pytest.mark.parametrize("repeat", range(TEST_SCALE))
@pytest.mark.parametrize("lba_count", [8, 8*1024, 0])  # 4K, 4M, all 
def test_trim_time_one_range(nvme0, nvme0n1, lba_count, repeat, qpair):
    buf = d.Buffer(4096)
    if lba_count == 0:
        lba_count = nvme0n1.id_data(7, 0)  # all lba
    buf.set_dsm_range(0, 0, lba_count)
    
    start_time = time.time()
    nvme0n1.dsm(qpair, buf, 1).waitdone()
    with open("report.csv", "a") as f:
        f.write('%.6f\n' % (time.time()-start_time))


@pytest.mark.parametrize("repeat", range(TEST_SCALE))
@pytest.mark.parametrize("io_size", [1, 8, 64, 512, 4096])  # 4K, 4M, all 
def test_trim_time_all_range_buffer(nvme0, nvme0n1, repeat, io_size, qpair):
    buf = d.Buffer(4096)
    for i in range(4096//16):
        buf.set_dsm_range(i, i*io_size, io_size)
    
    start_time = time.time()
    nvme0n1.dsm(qpair, buf, 1).waitdone()
    with open("report.csv", "a") as f:
        f.write('%.6f\n' % (time.time()-start_time))

    
# format
@pytest.mark.parametrize("repeat", range(TEST_SCALE))
def test_format_time(nvme0n1, repeat):
    start_time = time.time()
    nvme0n1.format()
    with open("report.csv", "a") as f:
        f.write('%.6f\n' % (time.time()-start_time))



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging
import nvme as d

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem


def test_pcie_pmcsr_d3hot(pcie, nvme0, buf):
    pm_offset = pcie.cap_offset(1)
    pmcs = pcie[pm_offset+4]
    logging.info("pmcs %x" % pmcs)
    nvme0.identify(buf).waitdone()

    # set d3hot
    pcie[pm_offset+4] = pmcs|3     #D3hot
    pmcs = pcie[pm_offset+4]
    logging.info("pmcs %x" % pmcs)

    # and exit d3hot
    time.sleep(1)
    pcie[pm_offset+4] = pmcs&0xfc  #D0
    pmcs = pcie[pm_offset+4]
    logging.info("pmcs %x" % pmcs)
    nvme0.identify(buf).waitdone()

    # set d3hot
    pcie[pm_offset+4] = pmcs|3     #D3hot
    pmcs = pcie[pm_offset+4]
    logging.info("pmcs %x" % pmcs)

    with pytest.raises(TimeoutError):
        with pytest.warns(UserWarning, match="ERROR status: 07/ff"):
            nvme0.identify(buf).waitdone()

    # and exit d3hot
    time.sleep(1)
    pcie[pm_offset+4] = pmcs&0xfc  #D0
    pmcs = pcie[pm_offset+4]
    logging.info("pmcs %x" % pmcs)
    nvme0.identify(buf).waitdone()

    
def test_pcie_capability_d3hot(pcie, nvme0n1):
    # get pm register
    assert None != pcie.cap_offset(1)
    pm_offset = pcie.cap_offset(1)
    pmcs = pcie[pm_offset+4]
    assert pcie.power_state == 0

    # set d3hot
    pcie.power_state = 3
    assert pcie.power_state == 3
    time.sleep(1)

    # and exit d3hot
    pcie.power_state = 0
    assert pcie.power_state == 0
    nvme0n1.ioworker(io_size=2, time=2).start().close()

    # again
    pcie.power_state = 0
    logging.info("curent power state: %d" % pcie.power_state)
    nvme0n1.ioworker(io_size=2, time=2).start().close()
    assert pcie.power_state == 0
    

def test_pcie_aspm_L1(pcie, nvme0, buf):
    #ASPM L1
    pcie.aspm = 2
    buf = d.Buffer(4096, 'controller identify data')
    nvme0.identify(buf, 0, 1).waitdone()
    time.sleep(1)
    #ASPM L0
    pcie.aspm = 0
    logging.info("model number: %s" % nvme0.id_data(63, 24, str))

def test_pcie_aspm_l1_and_d3hot(pcie, nvme0n1):
    pcie.aspm = 2
    assert pcie.aspm == 2
    pcie.power_state = 3
    time.sleep(1)
    pcie.power_state = 0
    pcie.aspm = 0
    assert pcie.aspm == 0
    nvme0n1.ioworker(io_size=2, time=2).start().close()

    pcie.power_state = 3
    assert pcie.power_state == 3
    pcie.aspm = 2
    time.sleep(1)
    pcie.aspm = 0
    assert pcie.aspm == 0
    pcie.power_state = 0
    assert pcie.power_state == 0
    nvme0n1.ioworker(io_size=2, time=2).start().close()
    assert pcie.aspm == 0

    pcie.power_state = 3
    assert pcie.power_state == 3
    time.sleep(1)
    pcie.power_state = 0
    assert pcie.power_state == 0
    nvme0n1.ioworker(io_size=2, time=2).start().close()


def test_pcie_ioworker_aspm(pcie, nvme0, buf):
    region_end = 256*1000*1000  # 1GB
    qdepth = min(1024, 1+(nvme0.cap&0xffff))
    
    # get the unsafe shutdown count
    def power_cycle_count():
        buf = d.Buffer(4096)
        nvme0.getlogpage(2, buf, 512).waitdone()
        return buf.data(115, 112)
    
    # run the test one by one
    subsystem = d.Subsystem(nvme0)
    nvme0n1 = d.Namespace(nvme0, 1, region_end)
    orig_unsafe_count = power_cycle_count()
    logging.info("power cycle count: %d" % orig_unsafe_count)

    # 128K random write, changing aspm
    cmdlog_list = [None]*1000
    with nvme0n1.ioworker(io_size=256,
                          lba_random=True,
                          read_percentage=30,
                          region_end=region_end,
                          time=20,
                          qdepth=qdepth, 
                          output_cmdlog_list=cmdlog_list):
        for i in range(10):
            time.sleep(1)
            pcie.aspm = 2
            time.sleep(1)
            pcie.aspm = 0
        
    # verify data in cmdlog_list
    time.sleep(5)
    assert True == nvme0n1.verify_enable(True)
    logging.info(cmdlog_list[-10:])
    read_buf = d.Buffer(256*512)
    qpair = d.Qpair(nvme0, 10)
    for cmd in cmdlog_list:
        slba = cmd[0]
        nlba = cmd[1]
        op = cmd[2]
        if nlba:
            def read_cb(cdw0, status1):
                nonlocal _slba
                if status1>>1:
                    logging.info("slba %d, 0x%x, _slba 0x%x, status 0x%x" % \
                                 (slba, slba, _slba, status1>>1))
                    
            logging.debug("verify slba %d, nlba %d" % (slba, nlba))
            _nlba = nlba//16
            for i in range(16):
                _slba = slba+i*_nlba
                nvme0n1.read(qpair, read_buf, _slba, _nlba, cb=read_cb).waitdone()
            
            # re-write to clear CRC mismatch
            nvme0n1.write(qpair, read_buf, slba, nlba, cb=read_cb).waitdone()
    qpair.delete()
    nvme0n1.close()

    # verify unsafe shutdown count
    unsafe_count = power_cycle_count()
    logging.info("power cycle count: %d" % unsafe_count)
    assert unsafe_count == orig_unsafe_count
    time.sleep(5)
    pcie.aspm = 0 




#
#  BSD LICENSE
#
#  Copyright (c) Crane Chu <cranechu@gmail.com>
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#    * Neither the name of Intel Corporation nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# -*- coding: utf-8 -*-


import pytest
import nvme as d

import time
import logging


rand = True
seq = False
TEST_SCALE = 100    #100, 10


def do_fill_drive(rand, nvme0n1, region_end, nvme0):
    io_size = 8 if rand else 128
    ns_size = nvme0n1.id_data(7, 0)
    io_count = ns_size//io_size
    io_per_second = []
    
    w = nvme0n1.ioworker(io_size=io_size, lba_align=io_size,
                     lba_random=rand, qdepth=512,
                     region_end=region_end,
                     io_count=io_count, read_percentage=0,
                     output_io_per_second=io_per_second).start()
    while w.running:
        nvme0.getfeatures(7).waitdone()
        time.sleep(1)
    r = w.close()
    logging.info(io_per_second)

    
# full drive seq write
def test_clean_and_fill_drive(nvme0, nvme0n1):
    nvme0n1.format(512)
    do_fill_drive(seq, nvme0n1, nvme0n1.id_data(7, 0), nvme0)


# random write 20% region to steady
def test_random_write_small_region(nvme0, nvme0n1):
    do_fill_drive(rand, nvme0n1, nvme0n1.id_data(7, 0)//5, nvme0)

    
# iops: read/write/mixed
# latency: read/write/mixed
@pytest.mark.parametrize("readp", [100, 90, 67, 50, 33, 10, 0])
def test_steady_iops_latency(nvme0, nvme0n1, readp):
    percentile_latency = dict.fromkeys([50, 90, 99, 99.9, 99.999])
    io_per_second = []

    w = nvme0n1.ioworker(io_size=8, lba_align=8,
                         lba_random=rand, qdepth=64,
                         region_end=nvme0n1.id_data(7, 0)//5, 
                         time=TEST_SCALE, read_percentage=readp,
                         output_percentile_latency=percentile_latency, 
                         output_io_per_second=io_per_second).start()
    while w.running:
        nvme0.getfeatures(7).waitdone()
        time.sleep(1)
    r = w.close()
    logging.info(io_per_second)
    consistency = w.iops_consistency(90)
    logging.info(percentile_latency)
    max_iops = (r.io_count_read+r.io_count_nonread)*1000//r.mseconds
    logging.info(max_iops)

    with open("report.csv", "a") as f:
        f.write('%d\n' % readp)
        f.write('%.2f\n' % (consistency/100))
        f.write('%d\n' % percentile_latency[99.9])
        f.write('%d\n' % max_iops)
        
    # latency against iops
    for iops_percentage in [20, 40, 60, 80, 100]:
        iops = max_iops*iops_percentage//100
        w = nvme0n1.ioworker(io_size=8, lba_align=8,
                             lba_random=rand, qdepth=64,
                             iops=iops, 
                             region_end=nvme0n1.id_data(7, 0)//5, 
                             time=TEST_SCALE, read_percentage=readp,
                             output_percentile_latency=percentile_latency).start()
        while w.running:
            nvme0.getfeatures(7).waitdone()
            time.sleep(1)
        r = w.close()
        logging.info("iops %d, latency %dus" % (iops, r.latency_average_us))

        # write to report
        with open("report.csv", "a") as f:
            f.write('%d\n' % iops)
            f.write('%d\n' % r.latency_average_us)



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


@pytest.fixture()
def cq(nvme0):
    ret = IOCQ(nvme0, 1, 10, PRP())
    yield ret
    ret.delete()

@pytest.fixture()
def sq(nvme0, cq):
    ret = IOSQ(nvme0, 1, 10, PRP(), cq.id)
    yield ret
    ret.delete()

    
def test_flush_with_read_write(nvme0, nvme0n1, qpair):
    # prepare data buffer and IO queue
    read_buf = Buffer(512)
    write_buf = Buffer(512)
    write_buf[10:21] = b'hello world'

    # send write and read command
    nvme0n1.write(qpair, write_buf, 0, 1).waitdone()
    nvme0n1.flush(qpair).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] == b'hello world'


def test_flush_invalid_nsid(nvme0, nvme0n1, cq, sq):
    # first cmd, invalid namespace
    cmd = SQE(0, 0xff)
    sq[0] = cmd
    sq.tail = 1
    time.sleep(0.1)
    status = (cq[0][3]>>17)&0x7ff
    assert status == 0x000b

    
def test_flush_all_namespace(nvme0, nvme0n1, cq, sq):
    # first cmd, all namespace
    cmd = SQE(0, 0xffffffff)
    sq[0] = cmd
    sq.tail = 1
    time.sleep(0.1)
    status = (cq[0][3]>>17)&0x7ff
    
    # NVMe1.4 page258: If bits 2:1 are set to 11b in the VWC field (refer to Figure 247) and the specified NSID is FFFFFFFFh, then
    # the Flush command applies to all namespaces attached to the controller processing the Flush command.
    # If bits 2:1 are set to 10b in the VWC field and the specified NSID is FFFFFFFFh, then the controller fails
    # the command with status code Invalid Namespace or Format. If bits 2:1 are cleared to 00b in the VWC field,
    # then the controller behavior if the specified NSID is FFFFFFFFh is not indicated. 
    vwc = nvme0.id_data(525)
    if (vwc>>1) == 3:   
        assert status == 0
    elif (vwc>>1) == 2:
        assert status == 0x000b
    else:
        pass

def test_flush_vwc_check(nvme0):
    # NVMe1.4 page258: Controllers compliant with
    # versions 1.4 and later of this specification shall not set bits 2:1 in the VWC field to the value of 00b.
    vwc = nvme0.id_data(525)
    vs = nvme0[8]
    logging.info("%d" % vs)
    if vs >= 0x010400:
        assert (vwc>>1) != 0
    else:
        pass



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


def test_apst_enabled(nvme0):
    if not nvme0.id_data(265):
        pytest.skip("APST is not enabled")

    pass


def test_host_controlled_thermal_management_enabled(nvme0):
    if not nvme0.id_data(322, 323):
        pytest.skip("APST is not enabled")

    pass


@pytest.mark.parametrize("ps", [4, 3, 2, 1, 0])
def test_format_at_power_state(nvme0, nvme0n1, ps):
    nvme0.setfeatures(0x2, cdw11=ps).waitdone()
    assert nvme0n1.format(ses=0) == 0
    assert nvme0n1.format(ses=1) == 0
    p = nvme0.getfeatures(0x2).waitdone()
    assert p == ps


def read_write_8M(nvme0n1, qpair):
    buf128k = Buffer(128*1024)
    lba128k = 128*1024//512

    # write all data
    slba = 0x10000
    for i in range(64):
        nvme0n1.write(qpair, buf128k, slba, lba128k)
        slba += lba128k
    qpair.waitdone(64)

    # read all data
    slba = 0x10000
    for i in range(64):
        nvme0n1.read(qpair, buf128k, slba, lba128k)
        slba += lba128k
    qpair.waitdone(64)
            
    
@pytest.mark.parametrize("ps_from", [0, 1, 2, 3, 4])
@pytest.mark.parametrize("ps_to", [0, 1, 2, 3, 4])
def test_power_state_transition(pcie, nvme0, nvme0n1, buf, qpair, ps_from, ps_to):
    # for accurate sleep delay
    import ctypes
    libc = ctypes.CDLL('libc.so.6')

    # enable ASPM and get original power state
    pcie.aspm = 2
    orig_ps = nvme0.getfeatures(0x2).waitdone()

    # disable apst
    nvme0.setfeatures(0xc, buf=buf).waitdone()

    # write data to LBA 0x5a
    buf = Buffer(512, ptype=32, pvalue=0x5a5a5a5a)
    nvme0n1.write(qpair, buf, 0x5a).waitdone()

    # test with delay 1us-1ms
    for delay in range(1000):
        # set beginning state
        nvme0.setfeatures(0x2, cdw11=ps_from).waitdone()
        libc.usleep(1000)

        # write and read data to invalidate cache
        read_write_8M(nvme0n1, qpair)

        # set end state
        nvme0.setfeatures(0x2, cdw11=ps_to)
        libc.usleep(delay)

        # read lba 0x5a and verify data
        nvme0n1.read(qpair, buf, 0x5a).waitdone()
        assert buf[0] == 0x5a
        for i in range(8, 512-8):
            assert buf[i] == 0x5a

        # consume the cpl of setfeatures above
        nvme0.waitdone()  # for setfeautres above

    # recover to original power state
    pcie.aspm = 0
    nvme0.setfeatures(0x2, cdw11=orig_ps).waitdone()


def test_power_state_transition_0_3_4(pcie, nvme0, nvme0n1, qpair, buf):
    # for accurate sleep delay
    import ctypes
    libc = ctypes.CDLL('libc.so.6')

    # enable ASPM and get original power state
    pcie.aspm = 2
    orig_ps = nvme0.getfeatures(0x2).waitdone()

    # disable apst
    nvme0.setfeatures(0xc, buf=buf).waitdone()

    # write data to LBA 0x5a
    buf = Buffer(512, ptype=32, pvalue=0x5a5a5a5a)
    nvme0n1.write(qpair, buf, 0x5a).waitdone()
    
    # test with delay 1us-10ms, 100 times each us
    for delay in range(10000):
        # write and read data to invalidate cache
        read_write_8M(nvme0n1, qpair)
        
        # PS0
        nvme0.setfeatures(0x2, cdw11=0).waitdone()
        libc.usleep(1000)
        
        # PS3, read, and PS4
        nvme0.setfeatures(0x2, cdw11=3)
        libc.usleep(delay)
        nvme0n1.read(qpair, buf, 0x5a)
        nvme0.setfeatures(0x2, cdw11=4)

        # check read result
        qpair.waitdone()
        assert buf[0] == 0x5a
        for i in range(8, 512-8):
            assert buf[i] == 0x5a

        # consume the cpl of setfeatures above
        nvme0.waitdone(2)

    # recover to original power state
    pcie.aspm = 0
    nvme0.setfeatures(0x2, cdw11=orig_ps).waitdone()


def test_power_state_async_with_io(pcie, nvme0, nvme0n1, buf, verify, duration=100):
    # for accurate sleep delay
    import ctypes
    libc = ctypes.CDLL('libc.so.6')

    # enable ASPM and get original power state
    pcie.aspm = 2
    orig_ps = nvme0.getfeatures(0x2).waitdone()

    # disable apst
    nvme0.setfeatures(0xc, buf=buf).waitdone()

    # start with PS0
    nvme0.setfeatures(0x2, cdw11=0).waitdone()
    
    # fill data for verify
    nvme0n1.ioworker(io_size=8,
                     lba_random=False,
                     read_percentage=0, 
                     region_end=1024*1024,
                     io_count=1024*1024//8).start().close()

    #mix read and PS setting
    w = nvme0n1.ioworker(io_size=8,
                         lba_random=True,
                         read_percentage=100, 
                         region_end=1024*1024,
                         iops=1000, 
                         time=duration).start()
    while w.running:
        nvme0.setfeatures(0x2, cdw11=3).waitdone()
        libc.usleep(999)
        nvme0.setfeatures(0x2, cdw11=4).waitdone()
        libc.usleep(999)
    w.close()



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

import nvme as d


@pytest.fixture(scope="function")
def ncqa(nvme0):
    num_of_queue = 0

    def test_greater_id(cdw0, status):
        nonlocal num_of_queue
        num_of_queue = 1+(cdw0 & 0xffff)
    nvme0.getfeatures(7, cb=test_greater_id).waitdone()
    logging.info("number of queue: %d" % num_of_queue)
    return num_of_queue


def get_aggregation_time_threshold(nvme0):
    time_threhold = 0

    def get_feature_cb(cdw0, status):
        nonlocal time_threhold
        time_threhold = cdw0 & 0xffff
    nvme0.getfeatures(8, cb=get_feature_cb).waitdone()
    logging.info("aggregation_time_threshold: 0x%x" % time_threhold)
    return time_threhold


def test_io_qpair_msix_interrupt_all(nvme0, nvme0n1, ncqa):
    buf = d.Buffer(4096)
    ql = []
    for i in range(ncqa):
        q = d.Qpair(nvme0, 8)
        ql.append(q)
        logging.info("qpair %d" % q.sqid)

        q.msix_clear()
        assert not q.msix_isset()
        nvme0n1.read(q, buf, 0, 8)
        time.sleep(0.1)
        assert q.msix_isset()
        q.waitdone()
    for q in ql:
        q.delete()


def test_io_qpair_msix_interrupt_mask(nvme0, nvme0n1, buf):
    q = d.Qpair(nvme0, 8)

    q.msix_clear()
    assert not q.msix_isset()
    nvme0n1.read(q, buf, 0, 8)
    time.sleep(1)
    assert q.msix_isset()
    q.waitdone()

    q.msix_clear()
    assert not q.msix_isset()
    nvme0n1.read(q, buf, 0, 8)
    time.sleep(1)
    assert q.msix_isset()
    q.waitdone()

    q.msix_clear()
    q.msix_mask()
    assert not q.msix_isset()
    nvme0n1.read(q, buf, 0, 8)
    assert not q.msix_isset()
    time.sleep(1)
    assert not q.msix_isset()
    q.msix_unmask()
    time.sleep(1)
    assert q.msix_isset()
    q.waitdone()

    q2 = d.Qpair(nvme0, 8)

    q.msix_clear()
    q2.msix_clear()
    assert not q.msix_isset()
    assert not q2.msix_isset()
    nvme0n1.read(q2, buf, 0, 8)
    time.sleep(1)
    assert not q.msix_isset()
    assert q2.msix_isset()
    q2.waitdone()

    q.delete()
    q2.delete()


def test_io_qpair_msix_interrupt_coalescing(nvme0, nvme0n1, buf, qpair):
    qpair.msix_clear()
    assert not qpair.msix_isset()

    # Get drvie interrupt aggregation time and threshold
    time_threhold = get_aggregation_time_threshold(nvme0)
    logging.info("interrupt aggregation time: %dus" %
                 ((time_threhold >> 8)*100))
    logging.info("interrupt aggregation threshold: %d" %
                 (time_threhold & 0xff))

    # 1 cmd, check interrupt latency
    nvme0n1.read(qpair, buf, 0, 8)
    start = time.time()
    while not qpair.msix_isset():
        pass
    latency1 = time.time()-start
    logging.info("interrupt latency %dus" % (latency1*1000000))
    qpair.waitdone()
    qpair.msix_clear()

    # aggregation time: 100*100us=0.01s, aggregation threshold: 2
    nvme0.setfeatures(8, cdw11=(200 << 8)+10).waitdone()

    # 2 cmd, check interrupt latency
    nvme0n1.read(qpair, buf, 0x8FFF, 8)
    nvme0n1.read(qpair, buf, 0x1600, 8)
    start = time.time()
    while not qpair.msix_isset():
        pass
    latency2 = time.time()-start
    logging.info("interrupt latency %dus" % (latency2*1000000))
    qpair.waitdone(2)
    qpair.msix_clear()

    # 1 cmd, check interrupt latency
    nvme0.setfeatures(8, cdw11=0).waitdone()
    nvme0n1.read(qpair, buf, 32, 8)
    start = time.time()
    while not qpair.msix_isset():
        pass
    latency1 = time.time()-start
    logging.info("interrupt latency %dus" % (latency1*1000000))
    qpair.waitdone()
    qpair.msix_clear()

    assert latency2 > latency1


def test_pcie_msix_cap_disable_ctrl(pcie, nvme0, nvme0n1, buf, qpair):
    msix_cap_addr = pcie.cap_offset(0x11)
    msix_ctrl = pcie.register(msix_cap_addr+2, 2)
    logging.info("msix_ctrl register [0x%x]= 0x%x" %
                 (msix_cap_addr+2, msix_ctrl))
    msix_table = pcie.register(msix_cap_addr+4, 4)
    logging.info("msix_table offset [0x%x]= 0x%x" %
                 (msix_cap_addr+4, msix_table))
    msix_pba = pcie.register(msix_cap_addr+8, 4)
    logging.info("msix_pba offset [0x%x]= 0x%x" % (msix_cap_addr+8, msix_pba))

    qpair.msix_clear()
    assert not qpair.msix_isset()
    nvme0n1.read(qpair, buf, 0, 8)
    time.sleep(0.1)
    assert qpair.msix_isset()
    qpair.waitdone()

    # Disable MSI-X bit
    qpair.msix_clear()
    pcie[msix_cap_addr+3] = (msix_ctrl >> 8) & 0x7F
    msix_ctrl = pcie.register(msix_cap_addr+2, 2)
    logging.info("After disable msi-x, msix_ctrl register [0x%x]= 0x%x" %
                 (msix_cap_addr+2, msix_ctrl))
    
    assert not qpair.msix_isset()
    nvme0n1.flush(qpair)
    time.sleep(1)
    assert not qpair.msix_isset()
    qpair.waitdone()

    # restore MSI-X bit
    pcie[msix_cap_addr+3] = (msix_ctrl >> 8) | 0x80
    msix_ctrl = pcie.register(msix_cap_addr+2, 2)
    logging.info("restore msix_ctrl register [0x%x]= 0x%x" %
                 (msix_cap_addr+2, msix_ctrl))
    



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem


def test_identify_all_nsid(nvme0):
    buf = Buffer(4096)
    nvme0.identify(buf, nsid=0, cns=1).waitdone()
    nvme0.identify(buf, nsid=1, cns=0).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/0b"):
        nvme0.identify(buf, nsid=0xff, cns=0).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/0b"):
        nvme0.identify(buf, nsid=2, cns=0).waitdone()


def test_identify_namespace_id_list(nvme0):
    if not nvme0.id_data(257, 256) & 0x8:
        pytest.skip("namespace management is not supported")

    buf = Buffer(4096)
    nvme0.identify(buf, nsid=0xffffffff, cns=0x10).waitdone()
    nvme0.identify(buf, nsid=0, cns=0x10).waitdone()


def test_identify_active_controller_list(nvme0):
    if not nvme0.id_data(257, 256) & 0x8:
        pytest.skip("namespace management is not supported")

    buf = Buffer(4096)
    nvme0.identify(buf, nsid=1, cns=0x12).waitdone()


def test_identify_controller_list(nvme0):
    if not nvme0.id_data(257, 256) & 0x8:
        pytest.skip("namespace management is not supported")

    buf = Buffer(4096)
    nvme0.identify(buf, nsid=1, cns=0x13).waitdone()


def test_identify_global_namespace(nvme0):
    if not nvme0.id_data(257, 256) & 0x8:
        pytest.skip("namespace management is not supported")

    buf = Buffer(4096)
    nvme0.identify(buf, nsid=0xffffffff, cns=0).waitdone()


def test_identify_namespace(nvme0):
    if not nvme0.id_data(257, 256) & 0x8:
        pytest.skip("namespace management is not supported")

    buf = Buffer(4096)
    nvme0.identify(buf, nsid=0xffffffff, cns=0).waitdone()


def test_identify_namespace_id_list(nvme0, buf):
    nvme0.identify(buf, nsid=0, cns=2).waitdone()
    assert buf[0] == 1
    assert buf[16] == 0

    nvme0.identify(buf, nsid=1, cns=0).waitdone()
    print(buf.dump(64))
    assert buf[0] != 0
    for i in range(8):
        assert buf[i] == buf[i+8]


def test_identify_namespace_utilization(nvme0, nvme0n1, buf):
    q = Qpair(nvme0, 10)
    orig_nuse = nvme0n1.id_data(23, 16)
    logging.info(orig_nuse)

    # trim lba 0
    buf.set_dsm_range(0, 0, 8)
    nvme0n1.dsm(q, buf, 1).waitdone()
    nuse = nvme0n1.id_data(23, 16)
    logging.info(nuse)
    assert nuse <= orig_nuse

    # write lba 0
    nvme0n1.write(q, buf, 0, 8).waitdone()
    time.sleep(0.1)  # some drive may update it at background
    nuse = nvme0n1.id_data(23, 16)
    logging.info(nuse)
    assert nuse == orig_nuse


def test_identify_namespace_identification_descriptor(nvme0, buf):
    nvme0.identify(buf, nsid=1, cns=3).waitdone()
    logging.info(buf.dump(64))
    assert buf[0] != 0
    with pytest.warns(UserWarning, match="ERROR status: 00/0b"):
        nvme0.identify(buf, nsid=0, cns=3).waitdone()


def test_identify_reserved_cns(nvme0, buf):
    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        nvme0.identify(buf, nsid=0, cns=0xff).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        nvme0.identify(buf, nsid=1, cns=0xff).waitdone()

def test_identify_nsze_ncap_nuse(nvme0, nvme0n1, buf):
    nsze = nvme0n1.id_data(7, 0)
    ncap = nvme0n1.id_data(15, 8)
    nuse = nvme0n1.id_data(23, 16)
    logging.info("nsze=0x%x,ncap=0x%x,nuse=0x%x" % (nsze,ncap,nuse))

    # Namespace Size >= Namespace Capacity >= Namespace Utilization
    assert nsze >= ncap >= nuse

    #if ANA reporting supported and in inaccessible or Persistent Loss state, nuse=0
    ana = nvme0.id_data(76, 76)
    ana_reporting = ana & 0x8
    ANACAP = nvme0.id_data(343, 343)
    inaccessible_loss = ANACAP & 0xc
    if ana_reporting and inaccessible_loss:
        assert nuse == 0



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem


@pytest.mark.parametrize("repeat", range(32))
def test_deallocate_and_write(nvme0, nvme0n1, repeat, qpair, 
                              lba_start=1, lba_step=3, lba_count=3):
    if not nvme0n1.supports(0x9):
        pytest.skip("dsm is not supprted")

    buf = Buffer(4096)
    pattern = repeat + (repeat<<8) + (repeat<<16) + (repeat<<24)
    write_buf = Buffer(4096, "write", pattern, 32)
    read_buf = Buffer(4096, "read")
    
    buf.set_dsm_range(0, lba_start+repeat*lba_step, lba_count)
    nvme0n1.dsm(qpair, buf, 1).waitdone()
    nvme0n1.write(qpair, write_buf, lba_start+repeat*lba_step, lba_count).waitdone()
    nvme0n1.read(qpair, read_buf, lba_start+repeat*lba_step, lba_count).waitdone()
    for i in range(lba_count):
        assert read_buf[i*512 + 10] == repeat
        

@pytest.mark.parametrize("repeat", range(32))
def test_deallocate_and_read(nvme0, nvme0n1, repeat, qpair, 
                             lba_start=1, lba_step=3, lba_count=3):
    if not nvme0n1.supports(0x9):
        pytest.skip("dsm is not supprted")

    buf = Buffer(4096)
    read_buf = Buffer(4096, "read")
    
    buf.set_dsm_range(0, lba_start+repeat*lba_step, lba_count)
    nvme0n1.dsm(qpair, buf, 1).waitdone()
    nvme0n1.read(qpair, read_buf, lba_start+repeat*lba_step, lba_count).waitdone()


def test_deallocate_out_of_range(nvme0, nvme0n1, qpair): 
    if not nvme0n1.supports(0x9):
        pytest.skip("dsm is not supprted")

    ncap = nvme0n1.id_data(15, 8)
    buf = Buffer(4096)

    buf.set_dsm_range(0, ncap-1, 1)
    nvme0n1.dsm(qpair, buf, 1).waitdone()
    
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        buf.set_dsm_range(0, ncap, 1)
        nvme0n1.dsm(qpair, buf, 1).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        buf.set_dsm_range(0, ncap-1, 2)
        nvme0n1.dsm(qpair, buf, 1).waitdone()


def test_deallocate_nr_maximum(nvme0, nvme0n1, qpair): 
    if not nvme0n1.supports(0x9):
        pytest.skip("dsm is not supprted")

    buf = Buffer(4096)

    for i in range(256):
        buf.set_dsm_range(i, i, 1)
    nvme0n1.dsm(qpair, buf, 256).waitdone()

    with pytest.raises(IndexError):
        for i in range(257):
            buf.set_dsm_range(i, i, 1)
    nvme0n1.dsm(qpair, buf, 257).waitdone()
    

def test_deallocate_correct_range(nvme0, nvme0n1, qpair):
    if not nvme0n1.supports(0x9):
        pytest.skip("dsm is not supprted")

    buf = Buffer(4096)
    nvme0n1.write(qpair, buf, 1, 3).waitdone()
    
    buf.set_dsm_range(0, 2, 1)
    nvme0n1.dsm(qpair, buf, 1).waitdone()

    nvme0n1.read(qpair, buf, 1, 1).waitdone()
    logging.info(buf[0:4])
    assert buf[0] == 1
    nvme0n1.read(qpair, buf, 2, 1).waitdone()
    logging.info(buf[0:4])
    nvme0n1.read(qpair, buf, 3, 1).waitdone()
    logging.info(buf[0:4])
    assert buf[0] == 3

    
def test_deallocate_multiple_range(nvme0, nvme0n1, qpair): 
    if not nvme0n1.supports(0x9):
        pytest.skip("dsm is not supprted")

    buf = Buffer(4096)
    nvme0n1.write(qpair, buf, 1, 4).waitdone()
    
    buf.set_dsm_range(0, 1, 1)
    buf.set_dsm_range(1, 2, 1)
    buf.set_dsm_range(2, 3, 1)
    nvme0n1.dsm(qpair, buf, 3).waitdone()

    nvme0n1.read(qpair, buf, 1, 1).waitdone()
    logging.debug(buf[0:4])
    nvme0n1.read(qpair, buf, 2, 1).waitdone()
    logging.debug(buf[0:4])
    nvme0n1.read(qpair, buf, 3, 1).waitdone()
    logging.debug(buf[0:4])
    nvme0n1.read(qpair, buf, 4, 1).waitdone()
    logging.debug(buf[0:4])
    assert buf[0] == 4
    
    nvme0n1.write(qpair, buf, 1, 4).waitdone()
    nvme0n1.read(qpair, buf, 1, 1).waitdone()
    logging.debug(buf[0:4])
    assert buf[0] == 1
    nvme0n1.read(qpair, buf, 2, 1).waitdone()
    logging.debug(buf[0:4])
    assert buf[0] == 2
    nvme0n1.read(qpair, buf, 3, 1).waitdone()
    logging.debug(buf[0:4])
    assert buf[0] == 3
    nvme0n1.read(qpair, buf, 4, 1).waitdone()
    logging.debug(buf[0:4])
    assert buf[0] == 4



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

import nvme as d
import scripts.tcg as tcg


skip_tcg = True
pytestmark = pytest.mark.skipif(skip_tcg, reason="Skip TCG")


@pytest.fixture()
def comid(nvme0):
    return tcg.Response(nvme0).receive().level0_discovery()


def test_uct01_level0_discovery(nvme0):
    comid = tcg.Response(nvme0).receive().level0_discovery()
    assert comid >= 1


def test_uct02_properties(nvme0):
    host_properties = {
        b"MaxComPacketSize": 4096,
        b"MaxPacketSize": 4076,
        b"MaxIndTokenSize": 4040,
    }

    comid = tcg.Response(nvme0).receive().level0_discovery()
    tcg.Command(nvme0, comid).properties(host_properties).send()
    tcg.Response(nvme0, comid).receive()


@pytest.mark.parametrize("new_passwd",
                         [b'123456',
                          b'1',
                          b'1234',
                          b'abcd',
                          b'123456789abcdef',
                          b'123456789abcdef0',
                          b'123456789abcdef0123456789abcdef0'])
def test_uct03_take_ownership(nvme0, comid, new_passwd):
    # take ownership
    tcg.Command(nvme0, comid).start_anybody_adminsp_session(0x65).send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).get_msid_cpin_pin(hsn, tsn).send()
    password = tcg.Response(nvme0, comid).receive().get_c_pin_msid()
    tcg.Command(nvme0, comid).end_session(hsn, tsn).send(False)
    tcg.Response(nvme0, comid).receive()

    tcg.Command(nvme0, comid).start_adminsp_session(0x66, password).send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).set_sid_cpin_pin(hsn, tsn, new_passwd).send()
    tcg.Response(nvme0, comid).receive()
    tcg.Command(nvme0, comid).end_session(hsn, tsn).send(False)
    tcg.Response(nvme0, comid).receive()

    # revert
    orig_timeout = nvme0.timeout
    nvme0.timeout = 100000
    tcg.Command(nvme0, comid).start_adminsp_session(0x69, new_passwd).send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).revert_tper(hsn, tsn).send()
    tcg.Response(nvme0, comid).receive()
    nvme0.timeout = orig_timeout


def test_uct04_activate_locking_sp(nvme0, comid, new_passwd=b'123456'):
    # take ownership
    tcg.Command(nvme0, comid).start_anybody_adminsp_session(0x65).send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).get_msid_cpin_pin(hsn, tsn).send()
    password = tcg.Response(nvme0, comid).receive().get_c_pin_msid()
    tcg.Command(nvme0, comid).end_session(hsn, tsn).send(False)
    tcg.Response(nvme0, comid).receive()

    tcg.Command(nvme0, comid).start_adminsp_session(0x66, password).send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).set_sid_cpin_pin(hsn, tsn, new_passwd).send()
    tcg.Response(nvme0, comid).receive()
    tcg.Command(nvme0, comid).end_session(hsn, tsn).send(False)
    tcg.Response(nvme0, comid).receive()

    # activate locking sp
    tcg.Command(nvme0, comid).start_adminsp_session(0x66, new_passwd).send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).get_locking_sp_lifecycle(hsn, tsn).send()
    tcg.Response(nvme0, comid).receive()
    tcg.Command(nvme0, comid).activate(hsn, tsn).send()
    tcg.Response(nvme0, comid).receive()
    tcg.Command(nvme0, comid).end_session(hsn, tsn).send(False)
    tcg.Response(nvme0, comid).receive()


def test_uct05_configuring_authorities(nvme0, comid, new_passwd=b'123456'):
    # enable user1 and passwd
    tcg.Command(nvme0, comid).start_auth_session(0x66, 0, new_passwd).send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).enable_user(hsn, tsn, 1).send()
    tcg.Response(nvme0, comid).receive()
    tcg.Command(nvme0, comid).set_new_passwd(hsn, tsn, 1, b"111111").send()
    tcg.Response(nvme0, comid).receive()
    tcg.Command(nvme0, comid).end_session(hsn, tsn).send(False)
    tcg.Response(nvme0, comid).receive()

    # admin auth session
    tcg.Command(nvme0, comid).start_auth_session(0x66, 0, new_passwd).send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).end_session(hsn, tsn).send(False)
    tcg.Response(nvme0, comid).receive()

    # user auth session and change passwd
    tcg.Command(nvme0, comid).start_auth_session(0x66, 1, b"111111").send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).set_new_passwd(hsn, tsn, 1, b"111111").send()
    tcg.Response(nvme0, comid).receive()
    tcg.Command(nvme0, comid).end_session(hsn, tsn).send(False)
    tcg.Response(nvme0, comid).receive()


def test_uct06_configuring_locking_objects(nvme0, nvme0n1, qpair, subsystem, verify, comid, new_passwd=b'123456'):
    # setup range and link to user1
    tcg.Command(nvme0, comid).start_auth_session(0x69, 0, new_passwd).send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).setup_range(hsn, tsn, 1, 0, 64).send()
    tcg.Response(nvme0, comid).receive()
    tcg.Command(nvme0, comid).lock_unlock_range(hsn, tsn, 1, False, False).send()
    tcg.Response(nvme0, comid).receive()
    tcg.Command(nvme0, comid).end_session(hsn, tsn).send(False)
    tcg.Response(nvme0, comid).receive()

    # write and verify
    buf = d.Buffer(64*512, pvalue=0x5aa55aa5, ptype=32)
    nvme0n1.write(qpair, buf, 0, 64).waitdone()
    buf = d.Buffer(64*512)
    nvme0n1.read(qpair, buf, 0, 64).waitdone()
    assert buf.data(11, 8) == 0x5aa55aa5

    
@pytest.mark.skip(reason="subsystem")
def test_uct06_configuring_locking_objects_powercycle(nvme0, nvme0n1, subsystem, verify, comid, new_passwd=b'123456'):    
    subsystem.poweroff()
    time.sleep(5)
    subsystem.poweron()
    nvme0.reset()
    
    qpair = d.Qpair(nvme0, 10)
    with pytest.warns(UserWarning, match="ERROR status: 02/86"):
        nvme0n1.read(qpair, buf, 0, 64).waitdone()
    qpair.delete()
    

def test_uct07_unlocking_range(nvme0, nvme0n1, qpair, verify, comid, new_passwd=b'123456'):
    tcg.Command(nvme0, comid).start_auth_session(0x69, 0, new_passwd).send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).lock_unlock_range(hsn, tsn, 1, False, False).send()
    tcg.Response(nvme0, comid).receive()
    tcg.Command(nvme0, comid).end_session(hsn, tsn).send(False)
    tcg.Response(nvme0, comid).receive()

    buf = d.Buffer(64*512)
    nvme0n1.read(qpair, buf, 0, 64).waitdone()
    assert buf.data(11, 8) == 0x5aa55aa5


def test_uct08_erasing_range(nvme0, nvme0n1, qpair, verify, comid, new_passwd=b'123456'):
    # TODO: skip on pyrite

    # verify data before erasing
    buf = d.Buffer(64*512)
    nvme0n1.read(qpair, buf, 0, 64).waitdone()
    assert buf.data(11, 8) == 0x5aa55aa5

    # erasing
    tcg.Command(nvme0, comid).start_auth_session(0x69, 0, new_passwd).send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).get_active_key(hsn, tsn, 1).send()
    prev_data = tcg.Response(nvme0, comid).receive().get_active_key()
    tcg.Command(nvme0, comid).gen_new_key(hsn, tsn, 1, prev_data).send()
    tcg.Response(nvme0, comid).receive()
    tcg.Command(nvme0, comid).end_session(hsn, tsn).send(False)
    tcg.Response(nvme0, comid).receive()

    # verify
    buf = d.Buffer(64*512)
    nvme0n1.read(qpair, buf, 0, 64).waitdone()
    assert buf.data(11, 8) == 0


def test_uct12_revert_locking_sp(nvme0, comid, new_passwd=b'123456'):
    # revert
    orig_timeout = nvme0.timeout
    nvme0.timeout = 100000
    tcg.Command(nvme0, comid).start_adminsp_session(0x69, new_passwd).send()
    hsn, tsn = tcg.Response(nvme0, comid).receive().start_session()
    tcg.Command(nvme0, comid).revert_tper(hsn, tsn).send()
    tcg.Response(nvme0, comid).receive()
    nvme0.timeout = orig_timeout



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem


def test_format_all_basic(nvme0, nvme0n1):
    if not nvme0.supports(0x80):
        pytest.skip("format is not support")

    orig_timeout = nvme0.timeout
    nvme0.timeout = 100000
    nvme0n1.format(512)
    
    nvme0.format(0).waitdone()
    nvme0.format(0, 0).waitdone()
    nvme0.format(0, 1).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 01/0a"):
        nvme0.format(2, 0).waitdone()

    nvme0.format(0, 0, 0xffffffff).waitdone()
    nvme0.format(0, 1, 0xffffffff).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/0b"):
        nvme0.format(0, 0, 0xfffffffb).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/0b"):
        nvme0.format(0, 1, 0xfffffffb).waitdone()

    nvme0n1.format(512)
    nvme0.timeout = orig_timeout
        

def test_format_verify_data(nvme0, nvme0n1, verify, qpair):
    if not nvme0.supports(0x80):
        pytest.skip("format is not support")

    orig_timeout = nvme0.timeout
    nvme0.timeout = 100000
        
    # prepare data buffer and IO queue
    read_buf = Buffer(4096)
    write_buf = Buffer(4096)
    write_buf[10:21] = b'hello world'

    nvme0n1.write(qpair, write_buf, 0, 1).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] == b'hello world'
    nvme0.format(0, 0).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] != b'hello world'
    assert read_buf[10:21] == b'\0'*11

    nvme0n1.write(qpair, write_buf, 0, 1).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] == b'hello world'
    nvme0.format(0, 1).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] != b'hello world'
    assert read_buf[10:21] == b'\0'*11

    nvme0n1.write(qpair, write_buf, 0, 1).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] == b'hello world'
    nvme0.format(0, 0, 0xffffffff).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] != b'hello world'
    assert read_buf[10:21] == b'\0'*11

    nvme0n1.write(qpair, write_buf, 0, 1).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] == b'hello world'
    nvme0.format(0, 1, 0xffffffff).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] != b'hello world'
    assert read_buf[10:21] == b'\0'*11

    # crypto erase
    fna = nvme0.id_data(524)
    if fna & 0x4:
        nvme0n1.write(qpair, write_buf, 0, 1).waitdone()
        nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
        assert read_buf[10:21] == b'hello world'
        nvme0.format(0, 2, 0xffffffff).waitdone()
        nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
        assert read_buf[10:21] != b'hello world'
        assert read_buf[10:21] == b'\0'*11

    nvme0n1.format(512)
    nvme0.timeout = orig_timeout
        

def test_format_invalid_ses(nvme0, nvme0n1, verify, qpair):
    if not nvme0.supports(0x80):
        pytest.skip("format is not support")

    orig_timeout = nvme0.timeout
    nvme0.timeout = 100000
        
    # prepare data buffer and IO queue
    read_buf = Buffer(4096)
    write_buf = Buffer(4096)
    write_buf[10:21] = b'hello world'

    nvme0n1.write(qpair, write_buf, 0, 1).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] == b'hello world'
    with pytest.warns(UserWarning, match="ERROR status: (00/02|01/0a)"):
        nvme0.format(0, 7, 0xffffffff).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] == b'hello world'
    
    nvme0n1.format(512)
    nvme0.timeout = orig_timeout

    
def test_format_invalid_lbaf(nvme0, nvme0n1, verify, qpair):
    if not nvme0.supports(0x80):
        pytest.skip("format is not support")

    orig_timeout = nvme0.timeout
    nvme0.timeout = 100000
        
    # prepare data buffer and IO queue
    read_buf = Buffer(4096)
    write_buf = Buffer(4096)
    write_buf[10:21] = b'hello world'

    nvme0n1.write(qpair, write_buf, 0, 1).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] == b'hello world'
    
    with pytest.warns(UserWarning, match="ERROR status: 01/0a"):
        nvme0.format(2).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] == b'hello world'        
    
    with pytest.warns(UserWarning, match="ERROR status: (00/02|01/0a)"):
        nvme0.format(0, 7).waitdone()
    nvme0n1.read(qpair, read_buf, 0, 1).waitdone()
    assert read_buf[10:21] == b'hello world'        

    nvme0n1.format(512)
    nvme0.timeout = orig_timeout
    



#
#  BSD LICENSE
#
#  Copyright (c) Crane Chu <cranechu@gmail.com>
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#    * Neither the name of Intel Corporation nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# -*- coding: utf-8 -*-


import pytest
import nvme as d

import os
import openpyxl

def test_generate_report_xlsx():
    wb = openpyxl.load_workbook(filename='scripts/performance/report_template.xlsx')
    sheet = wb['pynvme']
    line = 0

    # copy report.csv into excel file
    with open("report.csv", "r") as f:
        for value in f.readlines():
            line += 1
            value = value[:-1]
            try:
                value = float(value)
            except:
                pass

            # write to the cell of the excel file
            sheet['C%d' % line] = value

    # save the file
    wb.save(filename='performance_report.xlsx')
    os.remove("report.csv")



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


@pytest.fixture(scope="function")
def hmb(nvme0, buf):
    hmb_size = nvme0.id_data(275, 272)
    if hmb_size == 0:
        pytest.skip("hmb is not supported")
    
    hmb_buf = Buffer(4096*hmb_size)
    assert hmb_buf
    hmb_list_buf = Buffer(4096)
    assert hmb_list_buf
        
    hmb_list_buf[0:8] = hmb_buf.phys_addr.to_bytes(8, 'little')
    hmb_list_buf[8:12] = hmb_size.to_bytes(4, 'little')
    hmb_list_phys = hmb_list_buf.phys_addr

    # enable hmb
    nvme0.setfeatures(0x0d,
                      cdw11=1,
                      cdw12=hmb_size,
                      cdw13=hmb_list_phys&0xffffffff,
                      cdw14=hmb_list_phys>>32,
                      cdw15=1).waitdone()
    cdw0 = nvme0.getfeatures(0x0d, buf=buf).waitdone()
    assert cdw0 == 1
    assert buf.data(3, 0) == hmb_size
    assert buf.data(11, 4) == hmb_list_phys
    assert buf.data(15, 12) == 1
    logging.info("hmb enabled")
    yield

    # disable hmb
    nvme0.setfeatures(0x0d, cdw11=0).waitdone()
    cdw0 = nvme0.getfeatures(0x0d, buf=buf).waitdone()
    assert cdw0 == 0
    assert hmb_buf
    assert hmb_list_buf
    logging.info("hmb disabled")


def test_hmb_single_buffer(nvme0, nvme0n1, hmb):
    hmb_size = nvme0.id_data(275, 272)
    if hmb_size == 0:
        pytest.skip("hmb is not supported")

    # test with IO
    with nvme0n1.ioworker(io_size=8,
                          lba_random=False,
                          qdepth=8,
                          read_percentage=0,
                          time=3):
        pass
        
    with nvme0n1.ioworker(io_size=64,
                          lba_random=True,
                          read_percentage=0,
                          time=3):
        pass
        
        
def test_hmb_multiple_buffer(nvme0, nvme0n1, buf):
    hmb_size = nvme0.id_data(275, 272)
    if hmb_size == 0:
        pytest.skip("hmb is not supported")

    # allocate host memory
    logging.info(hmb_size)
    hmb_buf = Buffer(hmb_size*4096*2)
    hmb_buf_2 = Buffer(hmb_size*4096*2)  # reserve more buffer
    hmb_buf_3 = Buffer(hmb_size*4096*2)
    logging.info(hex(hmb_buf.phys_addr))
    logging.info(hex(hmb_buf_2.phys_addr))
    logging.info(hex(hmb_buf_3.phys_addr))
    del hmb_buf
    del hmb_buf_3

    # build the hmb list by 1MB chunks
    chunk_size = 0x100000
    chunk_per_page = chunk_size//4096
    hmb_list_buf = Buffer(4096)
    chunk_count = hmb_size*4096//chunk_size
    logging.info(chunk_count)
    addr = ((hmb_buf_2.phys_addr+0x200000)>>21)<<21  # align to 2MB
    for i in range(chunk_count):
        logging.info(hex(addr))
        hmb_list_buf[i*16:i*16+8] = addr.to_bytes(8, 'little')
        hmb_list_buf[i*16+8:i*16+12] = chunk_per_page.to_bytes(4, 'little')
        addr += chunk_size

    # enable hmb
    hmb_list_phys = hmb_list_buf.phys_addr
    logging.info(hex(hmb_list_phys))
    nvme0.setfeatures(0x0d,
                      cdw11=1,
                      cdw12=hmb_size,
                      cdw13=hmb_list_phys&0xffffffff,
                      cdw14=hmb_list_phys>>32,
                      cdw15=chunk_count).waitdone()

    cdw0 = nvme0.getfeatures(0x0d, buf=buf).waitdone()
    logging.info(cdw0)
    logging.info(buf.dump(16))
    assert cdw0 == 1
    assert buf.data(3, 0) == hmb_size
    assert buf.data(11, 4) == hmb_list_phys
    assert buf.data(15, 12) == chunk_count
    
    # test with IO
    for i in range(10):
        logging.info(i)
        with nvme0n1.ioworker(io_size=8,
                              lba_random=False,
                              qdepth=8,
                              read_percentage=0,
                              time=3):
            pass

    # disable hmb
    nvme0.setfeatures(0x0d, cdw11=0).waitdone()
    cdw0 = nvme0.getfeatures(0x0d, buf=buf).waitdone()
    logging.info(cdw0)
    logging.info(buf.dump(16))
    assert cdw0 == 0

    assert hmb_list_buf
    assert hmb_buf_2
    logging.info(hmb_list_buf.dump(64))
    logging.info(hmb_buf_2.dump(64))

    
def _test_reset_with_hmb_disabled(nvme0, nvme0n1, buf):
    hmb_size = nvme0.id_data(275, 272)
    if hmb_size == 0:
        pytest.skip("hmb is not supported")
        
    nvme0n1.format(512)

    # single host memory buffer
    hmb_buf = Buffer(4096*hmb_size+4096)
    assert hmb_buf
    hmb_list_buf = Buffer(4096)
    assert hmb_list_buf
    hmb_list_buf[0:8] = hmb_buf.phys_addr.to_bytes(8, 'little')
    hmb_list_buf[8:12] = hmb_size.to_bytes(4, 'little')
    hmb_list_phys = hmb_list_buf.phys_addr
    
    for i in range(3):
        logging.info(i)
        with nvme0n1.ioworker(io_size=8,
                              lba_random=False,
                              qdepth=8,
                              read_percentage=0,
                              time=30):
            time.sleep(5)
            
            # enable hmb    
            nvme0.setfeatures(0x0d,
                              cdw11=1,
                              cdw12=hmb_size,
                              cdw13=hmb_list_phys&0xffffffff,
                              cdw14=hmb_list_phys>>32,
                              cdw15=1).waitdone()
            time.sleep(5)

            #disable hmb
            nvme0.setfeatures(0x0d, cdw11=0).waitdone()
            time.sleep(5)
            
            nvme0.reset()

    logging.info(hmb_buf.dump(4096*2))
    del hmb_buf
    del hmb_list_buf
    

def test_multiple_hmb_buffer(nvme0, nvme0n1, buf):
    hmb_size = nvme0.id_data(275, 272)
    if hmb_size == 0:
        pytest.skip("hmb is not supported")

    return
    # single host memory buffer
    hmb_buffer_list = []
    hmb_list_buf = Buffer(hmb_size*16)
    assert hmb_list_buf
    for i in range(hmb_size):
        hmb_buf = Buffer()
        assert hmb_buf
        hmb_buffer_list.append(hmb_buf)
        hmb_list_buf[16*i+0 : 16*i+8] = hmb_buf.phys_addr.to_bytes(8, 'little')
        hmb_list_buf[16*i+8 : 16*i+12] = hmb_size.to_bytes(4, 'little')
    hmb_list_phys = hmb_list_buf.phys_addr

    # enable hmb    
    nvme0.setfeatures(0x0d,
                      cdw11=1,
                      cdw12=hmb_size,
                      cdw13=hmb_list_phys&0xffffffff,
                      cdw14=hmb_list_phys>>32,
                      cdw15=hmb_size).waitdone()
    
    for i in range(3):
        logging.info(i)
        with nvme0n1.ioworker(io_size=8,
                              lba_random=False,
                              qdepth=8,
                              read_percentage=0,
                              time=3):
            pass

    #disable hmb
    nvme0.setfeatures(0x0d, cdw11=0).waitdone()
    del hmb_buffer_list
    del hmb_list_buf
    
    



#
#  BSD LICENSE
#
#  Copyright (c) Crane Chu <cranechu@gmail.com>
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#    * Neither the name of Intel Corporation nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# -*- coding: utf-8 -*-


import pytest
import nvme as d

import time
import logging


TEST_SCALE = 10  #10, 1


def do_power_cycle(dirty, subsystem, nvme0n1, nvme0):
    if not dirty:
        # notify drive for a clean shutdown
        start_time = time.time()
        subsystem.shutdown_notify()
        logging.info("notify time %.6f sec" % (time.time()-start_time))

    # boot again
    csv_start = time.time()
    start_time = time.time()
    subsystem.power_cycle(10)
    nvme0.reset()
    logging.info("init time %.6f sec" % (time.time()-start_time-10))

    # first read time
    start_time = time.time()
    q = d.Qpair(nvme0, 16)
    b = d.Buffer(512)
    lba = nvme0n1.id_data(7, 0) - 1
    nvme0n1.read(q, b, lba).waitdone()
    logging.info("media ready time %.6f sec" % (time.time()-start_time))
    q.delete()
    
    # report to csv
    ready_time = time.time()-csv_start-10
    with open("report.csv", "a") as f:
        f.write('%.6f\n' % ready_time)

    
# rand write clean boot time
@pytest.mark.parametrize("repeat", range(TEST_SCALE))
@pytest.mark.parametrize("dirty", [False, True])
def test_boot_time_rand(nvme0, nvme0n1, subsystem, repeat, dirty):
    # write to make drive dirty
    with nvme0n1.ioworker(io_size=8, lba_align=8,
                          lba_random=True, qdepth=64,
                          read_percentage=0, time=TEST_SCALE):
        pass

    do_power_cycle(dirty, subsystem, nvme0n1, nvme0)

    
# seq write clean boot time
@pytest.mark.parametrize("repeat", range(TEST_SCALE))
@pytest.mark.parametrize("dirty", [False, True])
def test_boot_time_seq(nvme0, nvme0n1, subsystem, repeat, dirty):
    # write to make drive dirty
    with nvme0n1.ioworker(io_size=128, lba_align=128,
                          lba_random=False, qdepth=64,
                          read_percentage=0, time=TEST_SCALE):
        pass
    
    do_power_cycle(dirty, subsystem, nvme0n1, nvme0)



#
#  BSD LICENSE
#
#  Copyright (c) Crane Chu <cranechu@gmail.com>
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#    * Neither the name of Intel Corporation nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

import nvme as d


# intuitive, spec, qpair, vscode, debug, cmdlog, assert
def test_hello_world(nvme0, nvme0n1, qpair):
    # prepare data buffer and IO queue
    read_buf = d.Buffer()
    write_buf = d.Buffer()
    write_buf[10:21] = b'hello world'

    # send write and read command
    def write_cb(cdw0, status1):  # command callback function
        nvme0n1.read(qpair, read_buf, 0, 1)
    nvme0n1.write(qpair, write_buf, 0, 1, cb=write_cb)

    # wait commands complete and verify data
    assert read_buf[10:21] != b'hello world'
    qpair.waitdone(2)
    assert read_buf[10:21] == b'hello world'


# access PCIe/NVMe registers, identify data, pythonic
def test_registers_and_identify_data(pcie, nvme0, nvme0n1):
    logging.info("0x%x, 0x%x" % (pcie[0], pcie.register(0, 2)))
    logging.info("0x%08x, 0x%08x" % (nvme0[0], nvme0[4]))
    logging.info("model name: %s" % nvme0.id_data(63, 24, str))
    logging.info("vid: 0x%x" % nvme0.id_data(1, 0))
    logging.info("namespace size: %d" % nvme0n1.id_data(7, 0))


# Controller, sanitize, default parameters, getlogpage
def test_sanitize(nvme0: d.Controller, buf):
    if nvme0.id_data(331, 328) == 0:
        pytest.skip("sanitize operation is not supported")

    #import PySimpleGUI as sg

    logging.info("supported sanitize operation: %d" % nvme0.id_data(331, 328))
    #sg.OneLineProgressMeter('sanitize progress', 0, 100, 'progress', orientation='h')
    nvme0n1 = d.Namespace(nvme0, 1, 128*1000*1000//4)
    nvme0.sanitize().waitdone()  # sanitize clears namespace

    # check sanitize status in log page
    nvme0.getlogpage(0x81, buf, 20).waitdone()
    while buf.data(3, 2) & 0x7 != 1:  # sanitize operation is not completed
        time.sleep(1)
        nvme0.getlogpage(0x81, buf, 20).waitdone()
        progress = buf.data(1, 0)*100//0xffff
        #sg.OneLineProgressMeter('sanitize progress', progress, 100, 'progress', orientation='h')
        logging.info("%d%%" % progress)

    nvme0n1.close()


# simple ioworker, complicated io_size, python function call, CI
def test_ioworker_simplified(nvme0, nvme0n1: d.Namespace, qpair):
    nvme0n1.ioworker(time=1).start().close()
    nvme0n1.ioworker(io_size=[1, 2, 3, 7, 8, 16], time=1).start().close()
    nvme0n1.ioworker(op_percentage={2:10, 1:20, 0:30, 9:40}, time=1).start().close()
    test_hello_world(nvme0, nvme0n1, qpair)


# ioworker with admin commands, multiprocessing, log, cmdlog, pythonic
def subprocess_trim(pciaddr, seconds):
    pcie = d.Pcie(pciaddr)
    nvme0 = d.Controller(pcie, True)
    nvme0n1 = d.Namespace(nvme0)
    q = d.Qpair(nvme0, 8)
    buf = d.Buffer(4096)
    buf.set_dsm_range(0, 8, 8)

    # send trim commands
    start = time.time()
    while time.time()-start < seconds:
        nvme0n1.dsm(q, buf, 1).waitdone()

    q.delete()
    nvme0n1.close()
    pcie.close()

def test_ioworker_with_temperature_and_trim(nvme0, nvme0n1):
    test_seconds = 10

    # start trim process
    import multiprocessing
    mp = multiprocessing.get_context("spawn")
    p = mp.Process(target = subprocess_trim,
                   args = (nvme0.addr, test_seconds))
    p.start()

    # start read/write ioworker and admin commands
    smart_log = d.Buffer(512, "smart log")
    with nvme0n1.ioworker(io_size=256,
                          lba_random=False,
                          read_percentage=0,
                          time=test_seconds):
        for i in range(15):
            time.sleep(1)
            nvme0.getlogpage(0x02, smart_log, 512).waitdone()
            ktemp = smart_log.data(2, 1)

            from pytemperature import k2c
            logging.info("temperature: %0.2f degreeC" % k2c(ktemp))

    # wait trim process complete
    p.join()


# multiple ioworkers, PCIe, TCP, CPU, performance, ioworker return values
def test_multiple_controllers_and_namespaces(pciaddr):
    # address list of the devices to test
    addr_list = ['01:00.0', '02:00.0', '03:00.0', '04:00.0']
    addr_list = [pciaddr, ]
    test_seconds = 10

    # create all controllers and namespace
    pcie_list = [d.Pcie(a) for a in addr_list]
    nvme_list = [d.Controller(p) for p in pcie_list]
    ns_list = [d.Namespace(n) for n in nvme_list]

    # create two ioworkers on each namespace
    ioworkers = []
    for ns in ns_list:
        w = ns.ioworker(io_size=8, lba_align=8,
                        region_start=0, region_end=256*1024*8, # 1GB space
                        lba_random=False, qdepth=64,
                        read_percentage=100, time=test_seconds).start()
        ioworkers.append(w)
        w = ns.ioworker(io_size=8, lba_align=16,
                        region_start=256*1024*8, region_end=2*256*1024*8,
                        lba_random=True, qdepth=256,
                        read_percentage=0, time=test_seconds).start()
        ioworkers.append(w)

    # collect test results
    io_total = 0
    for w in ioworkers:
        r = w.close()
        io_total += (r.io_count_read+r.io_count_nonread)
    logging.info("total throughput: %d IOPS" % (io_total/test_seconds))

    for n in ns_list:
        n.close()

    for p in pcie_list:
        p.close()


# PCIe, different of power states and resets
def test_power_and_reset(pcie, nvme0, subsystem):
    pcie.aspm = 2              # ASPM L1
    pcie.power_state = 3       # PCI PM D3hot
    pcie.aspm = 0
    pcie.power_state = 0

    nvme0.reset()              # controller reset: CC.EN
    nvme0.getfeatures(7).waitdone()

    pcie.reset()               # PCIe reset: hot reset, TS1, TS2
    nvme0.reset()              # reset controller after pcie reset
    nvme0.getfeatures(7).waitdone()

    pcie.flr()                 # PCIe function level reset
    nvme0.reset()              # reset controller after pcie reset
    nvme0.getfeatures(7).waitdone()

    subsystem.reset()          # NVMe subsystem reset: NSSR
    nvme0.reset()              # controller reset: CC.EN
    nvme0.getfeatures(7).waitdone()

    subsystem.power_cycle(10)  # power cycle NVMe device: cold reset
    nvme0.reset()              # controller reset: CC.EN
    nvme0.getfeatures(7).waitdone()

    subsystem.poweroff()
    subsystem.poweron()
    nvme0.reset()              # controller reset: CC.EN
    nvme0.getfeatures(7).waitdone()


# test parameters, leverage innovations in python community
@pytest.mark.parametrize("io_count", [1, 9])
@pytest.mark.parametrize("lba_count", [1, 8])
@pytest.mark.parametrize("lba_offset", [0, 8])
def test_different_io_size_and_count(nvme0, nvme0n1, qpair,
                                     lba_offset, lba_count, io_count):
    # allcoate all DMA buffers for IO commands
    bufs = []
    for i in range(io_count):
        bufs.append(d.Buffer(lba_count*512))

    # send and reap all IO command dwords
    for i in range(io_count):
        nvme0n1.read(qpair, bufs[i], lba_offset, lba_count)
    qpair.waitdone(io_count)


# IO commands, fused operations, generic commands
def test_fused_operations(nvme0, nvme0n1):
    # create qpair and buffer for IO commands
    q = d.Qpair(nvme0, 10)
    b = d.Buffer()

    # separate compare and write commands
    nvme0n1.write(q, b, 8).waitdone()
    nvme0n1.compare(q, b, 8).waitdone()

    # implement fused compare and write operations with generic commands
    # Controller.send_cmd() sends admin commands,
    # and Namespace.send_cmd() here sends IO commands.
    nvme0n1.send_cmd(5|(1<<8), q, b, 1, 8, 0, 0)
    nvme0n1.send_cmd(1|(1<<9), q, b, 1, 8, 0, 0)
    q.waitdone(2)
    q.delete()


# protocol tests on queue, buffer, PRP, and doorbells
from psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE
def test_send_cmd_2sq_1cq(nvme0):
    # 2 SQ share one CQ
    cq = IOCQ(nvme0, 1, 10, PRP())
    sq1 = IOSQ(nvme0, 1, 10, PRP(), cqid=1)
    sq2 = IOSQ(nvme0, 2, 16, PRP(), cqid=1)

    # write lba0, 16K data organized by PRPList
    write_cmd = SQE(1, 1)       # write to namespace 1
    write_cmd.prp1 = PRP()      # PRP1 is a 4K page
    prp_list = PRPList()        # PRPList contains 3 pages
    prp_list[0] = PRP()
    prp_list[1] = PRP()
    prp_list[2] = PRP()
    write_cmd.prp2 = prp_list   # PRP2 points to the PRPList
    write_cmd[10] = 0           # starting LBA
    write_cmd[12] = 31          # LBA count: 32, 16K, 4 pages
    write_cmd.cid = 123         # verify cid later

    # send write commands in both SQ
    sq1[0] = write_cmd          # fill command dwords in SQ1
    write_cmd.cid = 567         # verify cid later
    sq2[0] = write_cmd          # fill command dwords in SQ2
    sq2.tail = 1                # ring doorbell of SQ2 first
    time.sleep(0.1)             # delay to ring SQ1,
    sq1.tail = 1                #  so command in SQ2 should complete first

    # wait for 2 command completions
    while CQE(cq[1]).p == 0: pass

    # check first cpl
    cqe = CQE(cq[0])
    assert cqe.sqid == 2
    assert cqe.sqhd == 1
    assert cqe.cid == 567

    # check second cpl
    cqe = CQE(cq[1])
    assert cqe.sqid == 1
    assert cqe.sqhd == 1
    assert cqe.cid == 123

    # update cq head doorbell to device
    cq.head = 2

    # delete all queues
    sq1.delete()
    sq2.delete()
    cq.delete()


def test_sanitize_operations_basic(nvme0, nvme0n1):  #L8
    if nvme0.id_data(331, 328) == 0:  #L9
        pytest.skip("sanitize operation is not supported")  #L10

    logging.info("supported sanitize operation: %d" % nvme0.id_data(331, 328))
    nvme0.sanitize().waitdone()  #L13

    # check sanitize status in log page
    buf = d.Buffer(4096)  #L16
    with pytest.warns(UserWarning, match="AER notification is triggered"):
        nvme0.getlogpage(0x81, buf, 20).waitdone()  #L17
        while buf.data(3, 2) & 0x7 != 1:  #L18
            time.sleep(1)
            nvme0.getlogpage(0x81, buf, 20).waitdone()  #L20
            progress = buf.data(1, 0)*100//0xffff
            logging.info("%d%%" % progress)


def test_buffer_read_write(nvme0, nvme0n1):
    buf = d.Buffer(512, 'ascii table')  #L2
    logging.info("physical address of buffer: 0x%lx" % buf.phys_addr)  #L3

    for i in range(512):
        buf[i] = i%256  #L6
    print(buf.dump(128))  #L7

    buf = d.Buffer(512, 'random', pvalue=100, ptype=0xbeef)  #L15
    print(buf.dump())
    buf = d.Buffer(512, 'random', pvalue=100, ptype=0xbeef)  #L17
    print(buf.dump())

    qpair = d.Qpair(nvme0, 10)
    nvme0n1.write(qpair, buf, 0).waitdone()
    nvme0n1.read(qpair, buf, 0).waitdone()
    print(buf.dump())
    qpair.delete()


@pytest.fixture()
def ncqa(nvme0):
    num_of_queue = 0
    def test_greater_id(cdw0, status):
        nonlocal num_of_queue
        num_of_queue = 1+(cdw0&0xffff)
    nvme0.getfeatures(7, cb=test_greater_id).waitdone()
    logging.info("number of queue: %d" % num_of_queue)
    return num_of_queue

def test_create_qpairs(nvme0, nvme0n1, buf, ncqa):
    qpair = d.Qpair(nvme0, 1024)
    nvme0n1.read(qpair, buf, 0)
    qpair.waitdone()
    nvme0n1.read(qpair, buf, 0, 8).waitdone()

    ql = []
    for i in range(ncqa-1):
        ql.append(d.Qpair(nvme0, 8))

    with pytest.raises(d.QpairCreationError):
        ql.append(d.Qpair(nvme0, 8))

    with pytest.warns(UserWarning, match="ioworker host ERROR -1: "):
        nvme0n1.ioworker(io_size=8, time=1000).start().close()

    qpair.delete()
    nvme0n1.ioworker(io_size=8, time=1).start().close()

    for q in ql:
        q.delete()


def test_namespace_multiple(pciaddr, buf):
    # create all controllers and namespace
    addr_list = [pciaddr, ] # add more DUT BDF here
    pcie_list = [d.Pcie(a) for a in addr_list]

    for p in pcie_list:
        nvmex = d.Controller(p)
        qpair = d.Qpair(nvmex, 8)
        nvmexn1 = d.Namespace(nvmex)

        #Check if support write uncorrectable command
        wuecc_support = nvmex.id_data(521, 520) & 0x2
        if wuecc_support != 0:
            nvmexn1.write_uncorrectable(qpair, 0, 8).waitdone()
            with pytest.warns(UserWarning, match="ERROR status: 02/81"):
                nvmexn1.read(qpair, buf, 0, 8).waitdone()

            nvmexn1.write(qpair, buf, 0, 8).waitdone()
            def this_read_cb(dword0, status1):
                assert status1>>1 == 0
                nvmexn1.write_uncorrectable(qpair, 0, 8)
            nvmexn1.read(qpair, buf, 0, 8, cb=this_read_cb).waitdone(2)

            def another_read_cb(dword0, status1):
                logging.info("dword0: 0x%08x" % dword0)
                logging.info("phase bit: %d" % (status1&1))
                logging.info("dnr: %d" % ((status1>>15)&1))
                logging.info("more: %d" % ((status1>>14)&1))
                logging.info("sct: 0x%x" % ((status1>>9)&0x7))
                logging.info("sc: 0x%x" % ((status1>>1)&0xff))
            with pytest.warns(UserWarning, match="ERROR status: 02/81"):
                nvmexn1.read(qpair, buf, 0, 8, cb=another_read_cb).waitdone()

        qpair.delete()
        nvmexn1.close()
        p.close()


@pytest.mark.parametrize("qcount", [1, 1, 2, 4])
def test_ioworker_iops_multiple_queue(nvme0n1, qcount):
    nvme0n1.format(512)

    l = []
    io_total = 0
    for i in range(qcount):
        a = nvme0n1.ioworker(io_size=8, lba_align=8,
                             region_start=0, region_end=256*1024*8, # 1GB space
                             lba_random=False, qdepth=16,
                             read_percentage=100, time=10).start()
        l.append(a)

    for a in l:
        r = a.close()
        io_total += r.io_count_read

    logging.info("Q %d IOPS: %.3fK" % (qcount, io_total/10000))


@pytest.mark.parametrize("iops", [100, 10*1000, 1000*1000])
def test_ioworker_fixed_iops(nvme0n1, iops):
    output_io_per_second = []
    nvme0n1.ioworker(io_size=8,
                     lba_random=True,
                     read_percentage=100,
                     iops=iops,
                     output_io_per_second=output_io_per_second,
                     time=10).start().close()
    logging.info(output_io_per_second)


def test_dsm_trim(nvme0: d.Controller, nvme0n1: d.Namespace, qpair: d.Qpair):
    trimbuf = d.Buffer(4096)

    # DUT info
    logging.info("model number: %s" % nvme0.id_data(63, 24, str))
    logging.info("firmware revision: %s" % nvme0.id_data(71, 64, str))

    # single range
    start_lba = 0
    lba_count = 8*1024
    trimbuf.set_dsm_range(0, start_lba, lba_count)
    nvme0n1.dsm(qpair, trimbuf, 1, attribute=0x4).waitdone()

    # multiple range
    lba_count = lba_count//256
    for i in range(256):
        trimbuf.set_dsm_range(i, start_lba+i*lba_count, lba_count)
    nvme0n1.dsm(qpair, trimbuf, 256).waitdone()


def test_ioworker_performance(nvme0n1):
    import matplotlib.pyplot as plt

    output_io_per_second = []
    percentile_latency = dict.fromkeys([90, 99, 99.9, 99.99, 99.999])
    nvme0n1.ioworker(io_size=8,
                     lba_random=True,
                     read_percentage=100,
                     output_io_per_second=output_io_per_second,
                     output_percentile_latency=percentile_latency,
                     time=10).start().close()
    logging.info(output_io_per_second)
    logging.info(percentile_latency)

    X = []
    Y = []
    for _, k in enumerate(percentile_latency):
        X.append(k)
        Y.append(percentile_latency[k])

    plt.plot(X, Y)
    plt.xscale('log')
    plt.yscale('log')
    #plt.show()


def test_ioworker_jedec_enterprise_workload(nvme0n1):
    distribution = [1000]*5 + [200]*15 + [25]*80
    iosz_distribution = {1: 4,
                         2: 1,
                         3: 1,
                         4: 1,
                         5: 1,
                         6: 1,
                         7: 1,
                         8: 67,
                         16: 10,
                         32: 7,
                         64: 3,
                         128: 3}

    nvme0n1.ioworker(io_size=iosz_distribution,
                     distribution = distribution,
                     lba_random=True,
                     read_percentage=0,
                     ptype=0xbeef, pvalue=100,
                     time=10).start().close()


def test_power_on_off(nvme0):
    def poweron():
        logging.info("poweron")
        pass
    def poweroff():
        logging.info("poweroff")
        pass
    subsystem = d.Subsystem(nvme0, poweron, poweroff)

    subsystem = d.Subsystem(nvme0)
    subsystem.poweroff()
    subsystem.poweron()
    nvme0.reset()


def test_init_nvme_customerized(pcie):
    def nvme_init(nvme0):
        logging.info("user defined nvme init")

        nvme0[0x14] = 0
        while not (nvme0[0x1c]&0x1) == 0: pass

        # 3. set admin queue registers
        nvme0.init_adminq()

        # 4. set register cc
        nvme0[0x14] = 0x00460000

        # 5. enable cc.en
        nvme0[0x14] = 0x00460001

        # 6. wait csts.rdy to 1
        while not (nvme0[0x1c]&0x1) == 1: pass

        # 7. identify controller
        nvme0.identify(d.Buffer(4096)).waitdone()

        # 8. create and identify all namespace
        nvme0.init_ns()

        # 9. set/get num of queues, 2 IO queues
        nvme0.setfeatures(0x7, cdw11=0x00010001).waitdone()
        nvme0.init_queues(nvme0.getfeatures(0x7).waitdone())

        # 10. send out all aer
        aerl = nvme0.id_data(259)+1
        for i in range(aerl):
            nvme0.aer()

    # 1. set pcie registers
    pcie.aspm = 0

    # 2. disable cc.en and wait csts.rdy to 0
    nvme0 = d.Controller(pcie, nvme_init_func=nvme_init)

    # test with ioworker
    nvme0n1 = d.Namespace(nvme0)
    qpair = d.Qpair(nvme0, 10)
    nvme0n1.ioworker(time=1).start().close()
    qpair2 = d.Qpair(nvme0, 10)
    with pytest.raises(d.QpairCreationError):
        qpair3 = d.Qpair(nvme0, 10)

    qpair.delete()
    qpair2.delete()
    nvme0n1.close()


def test_ioworker_op_dict_trim(nvme0n1):
    cmdlog_list = [None]*10000
    op_percentage = {2: 40, 9: 30, 1: 30}
    nvme0n1.ioworker(io_size=8,
                     io_count=10000,
                     op_percentage=op_percentage,
                     output_cmdlog_list=cmdlog_list).start().close()

    op_log = [c[2] for c in cmdlog_list]
    for op in (2, 9, 1):
        logging.info("occurance of %d: %d" % (op, op_log.count(op)))


def test_ioworker_io_sequence_read_write_trim_flush_uncorr(nvme0n1):
    cmd_seq = [(000000, 1, 0, 8),  #L2
               (200000, 2, 3, 1),
               (400000, 1, 2, 1),
               (600000, 9, 1, 1),
               (800000, 4, 0, 8),
               (1000000, 0, 0, 0)]
    cmdlog_list = [None]*len(cmd_seq)  #L8

    r = nvme0n1.ioworker(io_sequence=cmd_seq,  #L10
                         output_cmdlog_list=cmdlog_list).start().close()

    assert r.mseconds > 1000  #L13
    assert cmdlog_list[-1][2] == 0  #L14
    assert cmdlog_list[-2][2] == 4
    assert cmdlog_list[-3][2] == 9
    assert cmdlog_list[-4][2] == 1
    assert cmdlog_list[-5][2] == 2
    assert cmdlog_list[-6][2] == 1


def test_aer_with_multiple_sanitize(nvme0, nvme0n1, buf):  #L8
    if nvme0.id_data(331, 328) == 0:  #L9
        pytest.skip("sanitize operation is not supported")  #L10

    logging.info("supported sanitize operation: %d" % nvme0.id_data(331, 328))

    for i in range(3):
        nvme0.sanitize().waitdone()  #L13

        # check sanitize status in log page
        with pytest.warns(UserWarning, match="AER notification is triggered"):
            nvme0.getlogpage(0x81, buf, 20).waitdone()  #L17
            while buf.data(3, 2) & 0x7 != 1:  #L18
                time.sleep(1)
                nvme0.getlogpage(0x81, buf, 20).waitdone()  #L20
                progress = buf.data(1, 0)*100//0xffff
                logging.info("%d%%" % progress)


def test_verify_partial_namespace(nvme0):
    region_end=1024*1024*1024//512  # 1GB space
    nvme0n1 = d.Namespace(nvme0, 1, region_end)
    assert True == nvme0n1.verify_enable(True)

    nvme0n1.format()
    nvme0n1.ioworker(io_size=8,
                     lba_random=True,
                     region_end=region_end,
                     read_percentage=50,
                     time=30).start().close()
    nvme0n1.close()


def test_jsonrpc_list_qpairs(pciaddr):
    import json
    import socket

    # create the jsonrpc client
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.connect('/var/tmp/pynvme.sock')

    def jsonrpc_call(sock, method, params=[]):
        # create and send the command
        req = {}
        req['id'] = 1234567890
        req['jsonrpc'] = '2.0'
        req['method'] = method
        req['params'] = params
        sock.sendall(json.dumps(req).encode('ascii'))

        # receive the result
        resp = json.loads(sock.recv(4096).decode('ascii'))
        assert resp['id'] == 1234567890
        assert resp['jsonrpc'] == '2.0'
        return resp['result']

    result = jsonrpc_call(sock, 'list_all_qpair')
    assert len(result) == 0

    # create controller and admin queue
    pcie = d.Pcie(pciaddr)
    nvme0 = d.Controller(pcie)

    result = jsonrpc_call(sock, 'list_all_qpair')
    assert len(result) == 1
    assert result[0]['qid']-1 == 0

    result = jsonrpc_call(sock, 'list_all_qpair')
    assert len(result) == 1
    assert result[0]['qid']-1 == 0

    q1 = d.Qpair(nvme0, 8)
    result = jsonrpc_call(sock, 'list_all_qpair')
    assert len(result) == 2
    assert result[0]['qid']-1 == 0
    assert result[1]['qid']-1 == 1

    q2 = d.Qpair(nvme0, 8)
    result = jsonrpc_call(sock, 'list_all_qpair')
    assert len(result) == 3
    assert result[0]['qid']-1 == 0
    assert result[1]['qid']-1 == 1
    assert result[2]['qid']-1 == 2

    q1.delete()
    result = jsonrpc_call(sock, 'list_all_qpair')
    assert len(result) == 2
    assert result[0]['qid']-1 == 0
    assert result[1]['qid']-1 == 2

    q2.delete()
    result = jsonrpc_call(sock, 'list_all_qpair')
    assert len(result) == 1
    assert result[0]['qid']-1 == 0

    pcie.close()
    result = jsonrpc_call(sock, 'list_all_qpair')
    assert len(result) == 0


def test_powercycle_with_qpair(nvme0, nvme0n1, buf, subsystem):
    qpair = d.Qpair(nvme0, 16)
    nvme0n1.write(qpair, buf, 0).waitdone()
    nvme0n1.read(qpair, buf, 0).waitdone()

    # delete qpair before power cycle, and then reset controller, recreate qpair
    qpair.delete()
    subsystem.power_cycle(10)
    nvme0.reset()
    qpair = d.Qpair(nvme0, 16)

    nvme0n1.read(qpair, buf, 0).waitdone()
    qpair.delete()


def test_reset_time(pcie):
    def nvme_init(nvme0):
        logging.info("user defined nvme init")

        nvme0[0x14] = 0
        while not (nvme0[0x1c]&0x1) == 0: pass
        logging.info(time.time())

        # 3. set admin queue registers
        nvme0.init_adminq()
        logging.info(time.time())

        # 5. enable cc.en
        nvme0[0x14] = 0x00460001

        # 6. wait csts.rdy to 1
        while not (nvme0[0x1c]&0x1) == 1: pass
        logging.info(time.time())

        # 7. identify controller
        nvme0.identify(d.Buffer(4096)).waitdone()
        logging.info(time.time())

        nvme0.setfeatures(0x7, cdw11=0x00ff00ff).waitdone()
        nvme0.init_queues(nvme0.getfeatures(0x7).waitdone())

    logging.info("1: nvme init")
    logging.info(time.time())
    nvme0 = d.Controller(pcie, nvme_init_func=nvme_init)
    subsystem = d.Subsystem(nvme0)

    qpair = d.Qpair(nvme0, 10)
    qpair2 = d.Qpair(nvme0, 10)
    qpair3 = d.Qpair(nvme0, 10)
    qpair.delete()
    qpair2.delete()
    qpair3.delete()

    logging.info("2: nvme reset")
    logging.info(time.time())
    nvme0.reset()

    logging.info("3: power cycle")
    subsystem.poweroff()
    logging.info(time.time())
    subsystem.poweron()
    nvme0.reset()


@pytest.mark.parametrize("ps", range(1, 5, 1))
def test_power_state_transition(pcie, nvme0, nvme0n1, qpair, buf, ps):
    # for accurate sleep delay
    import ctypes
    libc = ctypes.CDLL('libc.so.6')

    # write data to LBA 0x5a
    nvme0n1.write(qpair, buf, 0x5a).waitdone()

    # enable ASPM and get original power state
    pcie.aspm = 2
    orig_ps = nvme0.getfeatures(0x2).waitdone()

    # test with delay 1us-1ms
    for i in range(1000):
        # fix on power state 0 before test
        nvme0.setfeatures(0x2, cdw11=0).waitdone()
        libc.usleep(1000)

        # change power status
        nvme0.setfeatures(0x2, cdw11=ps)
        libc.usleep(i)

        # read lba 0x5a and verify data
        nvme0n1.read(qpair, buf, 0x5a).waitdone()
        assert buf[0] == 0x5a

        # consume the cpl of setfeatures above
        nvme0.waitdone()  # for setfeautres above

    # recover to original power state
    pcie.aspm = 0
    nvme0.setfeatures(0x2, cdw11=orig_ps).waitdone()


@pytest.mark.parametrize("nsid", [0, 1, 0xffffffff])
def test_getlogpage_nsid(nvme0, buf, nsid):
    logging.info("model name: %s, nsid %d" % (nvme0.id_data(63, 24, str), nsid))
    nvme0.getlogpage(0xCA, buf, 512, nsid=nsid).waitdone()
    nvme0.getlogpage(0x02, buf, 512, nsid=nsid).waitdone()


def test_ioworker_with_temperature(nvme0, nvme0n1, buf):
    with nvme0n1.ioworker(io_size=256,
                          time=30,
                          op_percentage={0:10,  # flush
                                         2:60,  # read
                                         9:30}), \
         nvme0n1.ioworker(io_size=8,
                          time=30,
                          op_percentage={0:10,  # flush
                                         9:10,  # trim
                                         1:80}):# write
        for i in range(40):
            time.sleep(1)
            nvme0.getlogpage(0x02, buf, 512).waitdone()
            ktemp = buf.data(2, 1)
            from pytemperature import k2c
            logging.info("temperature: %0.2f degreeC" %
                         k2c(ktemp))


def test_ioworker_jedec_enterprise_workload_512(nvme0n1):
    distribution = [1000]*5 + [200]*15 + [25]*80
    iosz_distribution = {1: 4,
                         2: 1,
                         3: 1,
                         4: 1,
                         5: 1,
                         6: 1,
                         7: 1,
                         8: 67,
                         16: 10,
                         32: 7,
                         64: 3,
                         128: 3}

    output_percentile_latency = dict.fromkeys([99, 99.99, 99.9999])
    nvme0n1.ioworker(io_size=iosz_distribution,
                     lba_random=True,
                     qdepth=128,
                     distribution = distribution,
                     read_percentage=0,
                     ptype=0xbeef, pvalue=100,
                     time=30,
                     output_percentile_latency=\
                       output_percentile_latency).start().close()
    logging.info(output_percentile_latency)



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


def test_sq_cq_around(nvme0):
    cq = IOCQ(nvme0, 1, 3, PRP())
    sq = IOSQ(nvme0, 1, 5, PRP(), cqid=1)

    # send commands
    sq[0] = SQE(4<<16+0, 1)
    sq[1] = SQE(3<<16+0, 1)
    sq[2] = SQE(2<<16+0, 1)
    sq[3] = SQE(1<<16+0, 1)
    sq.tail = 4

    # check cq
    time.sleep(0.1)
    assert cq[0][3] == 0x10004
    assert cq[1][3] == 0x10003
    assert cq[2][3] == 0
    
    cq.head = 1
    time.sleep(0.1)
    assert cq[0][3] == 0x10004
    assert cq[2][3] == 0x10002
    
    cq.head = 2
    time.sleep(0.1)
    assert cq[2][3] == 0x10002
    assert cq[1][3] == 0x10003
    assert cq[0][3] == 0x00001

    sq.delete()
    cq.delete()


def test_sq_overflow(nvme0):
    cq = IOCQ(nvme0, 1, 5, PRP())
    sq = IOSQ(nvme0, 1, 2, PRP(), cqid=1)

    # send commands
    sq[0] = SQE(4<<16+0, 1); sq.tail = 1; time.sleep(0.1)
    sq[1] = SQE(3<<16+0, 1); sq.tail = 0; time.sleep(0.1)

    # check cq
    assert cq[0][3] == 0x10004
    assert cq[1][3] == 0x10003
    assert cq[2][3] == 0

    time.sleep(1)
    assert cq[2][3] == 0
    assert cq[3][3] == 0
    assert cq[4][3] == 0
    logging.debug(sq[0])
    logging.debug(cq[0])

    sq.delete()
    cq.delete()


def test_delete_cq_before_sq(nvme0):
    cq = IOCQ(nvme0, 1, 5, PRP())
    sq = IOSQ(nvme0, 1, 2, PRP(), cqid=1)
    with pytest.warns(UserWarning, match="ERROR status: 01/0c"):
        cq.delete()


def test_sq_doorbell(nvme0):
    cq = IOCQ(nvme0, 1, 5, PRP())
    sq = IOSQ(nvme0, 1, 2, PRP(), cqid=1)
    sq.tail = 1

    time.sleep(0.1)
    sq.delete()
    cq.delete()


@pytest.mark.parametrize("tail", [0, 2, 3, 0xffff, 0x10000])
def test_sq_doorbell_invalid(nvme0, tail):
    cq = IOCQ(nvme0, 1, 5, PRP())
    sq = IOSQ(nvme0, 1, 2, PRP(), cqid=1)
    sq.tail = tail

    time.sleep(0.1)
    #Invalid Doorbell Write Value
    with pytest.warns(UserWarning, match="AER notification is triggered: 0x10100"):
        nvme0.getfeatures(7).waitdone()
    sq.delete()
    cq.delete()


def test_cq_doorbell_valid(nvme0):
    cq = IOCQ(nvme0, 1, 5, PRP())
    time.sleep(0.1)
    cq.delete()


@pytest.mark.parametrize("head", [5, 6, 7, 0xffff])
def test_cq_doorbell_invalid(nvme0, head):
    cq = IOCQ(nvme0, 1, 5, PRP())
    sq = IOSQ(nvme0, 1, 2, PRP(), cqid=1)
    sq.tail = 1

    time.sleep(0.1)
    cq.head = head
    time.sleep(0.1)
    
    #Invalid Doorbell Write Value
    with pytest.warns(UserWarning, match="AER notification is triggered: 0x10100"):
        nvme0.getfeatures(7).waitdone()
    sq.delete()
    cq.delete()

    
def test_sq_cq_another_sq(nvme0):
    cq = IOCQ(nvme0, 1, 3, PRP())

    # send commands in sq1
    sq = IOSQ(nvme0, 1, 3, PRP(), cqid=1)
    sq[0] = SQE(4<<16+0, 1)
    sq[1] = SQE(3<<16+0, 1)
    sq.tail = 2

    sq2 = IOSQ(nvme0, 2, 3, PRP(), cqid=1)
    sq2[0] = SQE(2<<16+0, 1)
    sq2[1] = SQE(1<<16+0, 1)
    sq2.tail = 2

    # check cq
    time.sleep(0.1)
    assert cq[0][3] == 0x10004
    assert cq[1][3] == 0x10003
    assert cq[2][3] == 0
    
    cq.head = 1
    time.sleep(0.1)
    assert cq[2][3] == 0x10002
    assert cq[0][3] == 0x10004
    
    cq.head = 2
    time.sleep(0.1)
    assert cq[0][3] == 0x00001

    sq.delete()
    sq2.delete()
    cq.delete()


def _test_cq_sqhd_aer(nvme0,buf):
    #Create cq and sq
    a=()
    cq = IOCQ(nvme0, 1, 5, PRP())
    sq = IOSQ(nvme0, 1, 5, PRP(), cqid=1)

    #Trigger aer cmd response by invaild doorbell write value
    sq.tail = 5
    time.sleep(0.1)

    #Invalid Doorbell Write Value
    with pytest.warns(UserWarning, match="AER notification is triggered: 0x10100"):
        nvme0.waitdone()
    sq.delete()
    cq.delete()

    def call_back_cpl(cpl):
        nonlocal a;a=cpl

    #Check normal cmd cqe sqhd value
    nvme0.getlogpage(1, buf,cb=call_back_cpl).waitdone()
    sqhd1=a[2]&0xffff
    logging.info(sqhd1)

    nvme0.aer(cb=call_back_cpl)
    nvme0.getlogpage(1, buf,cb=call_back_cpl).waitdone()
    sqhd2=a[2]&0xffff
    logging.info(sqhd2)
    assert sqhd2==sqhd1+2

    #Trigger aer cmd response by invaild doorbell write value
    cq = IOCQ(nvme0, 1, 5, PRP())
    sq = IOSQ(nvme0, 1, 5, PRP(), cqid=1)
    sq.tail = 5
    time.sleep(0.1)

    #Invalid Doorbell Write Value
    with pytest.warns(UserWarning, match="AER notification is triggered: 0x10100"):
        nvme0.waitdone()
        #verify aer cmd cqe sqhd
        sqhd_aer=a[2]&0xffff
        sqhd_aer=sqhd2+2
        logging.info("aer sqhd is {}".format(sqhd_aer))

    sq.delete()
    cq.delete()

    #verify the following cmd cqe sqhd
    nvme0.getlogpage(1, buf,cb=call_back_cpl).waitdone()
    sqhd3=a[2]&0xffff
    assert sqhd3==sqhd_aer+3
    logging.info(sqhd3)





# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging
import warnings

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem


def test_sanitize_operations_basic(nvme0, nvme0n1, buf):
    if nvme0.id_data(331, 328) == 0:  #L9
        pytest.skip("sanitize operation is not supported")  #L10

    logging.info("supported sanitize operation: %d" % nvme0.id_data(331, 328))
    nvme0.sanitize().waitdone()  #L13

    # check sanitize status in log page
    with pytest.warns(UserWarning, match="AER notification is triggered"):
        nvme0.getlogpage(0x81, buf, 20).waitdone()  #L17
        while buf.data(3, 2) & 0x7 != 1:  #L18
            time.sleep(1)
            nvme0.getlogpage(0x81, buf, 20).waitdone()  #L20
            progress = buf.data(1, 0)*100//0xffff
            logging.info("%d%%" % progress)

    # check sanitize status
    nvme0.getlogpage(0x81, buf, 20).waitdone()
    assert buf.data(3, 2) & 0x7 == 1


def test_sanitize_operations_powercycle(nvme0, buf, subsystem):
    return
    if nvme0.id_data(331, 328) == 0:  #L9
        pytest.skip("sanitize operation is not supported")  #L10

    logging.info("supported sanitize operation: %d" % nvme0.id_data(331, 328))
    
    # slow down the sanitize
    nvme0.setfeatures(0x2, cdw11=2).waitdone()
    nvme0.sanitize().waitdone()  #L13
    
    subsystem.power_cycle()
    nvme0.reset()

    # check sanitize status in log page
    with pytest.warns(UserWarning, match="AER notification is triggered"):
        nvme0.getlogpage(0x81, buf, 20).waitdone()  #L17
        while buf.data(3, 2) & 0x7 != 1:  #L18
            time.sleep(1)
            nvme0.getlogpage(0x81, buf, 20).waitdone()  #L20
            progress = buf.data(1, 0)*100//0xffff
            logging.info("%d%%" % progress)
            
    # check sanitize status
    nvme0.getlogpage(0x81, buf, 20).waitdone()
    assert buf.data(3, 2) & 0x7 == 1

    # recover power mode
    nvme0.setfeatures(0x2, cdw11=0).waitdone()

    
def test_sanitize_operations_powercycle_post(nvme0, nvme0n1, buf, subsystem, qpair):
    return
    if nvme0.id_data(331, 328) == 0:  #L9
        pytest.skip("sanitize operation is not supported")  #L10
        
    logging.info("verify data after sanitize")
    nvme0n1.read(qpair, buf, 11, 1).waitdone()
    assert buf[0] == 0
    assert buf[511] == 0

    # read after sanitize
    nvme0n1.ioworker(io_size=8, lba_align=8,
                     region_start=0, region_end=256*1024*8, # 1GB space
                     lba_random=False, qdepth=16,
                     read_percentage=100, time=10).start().close()
    
    
def test_write_in_sanitize_operations(nvme0, nvme0n1, buf, qpair):
    if nvme0.id_data(331, 328) == 0:  #L9
        pytest.skip("sanitize operation is not supported")  #L10

    logging.info("supported sanitize operation: %d" % nvme0.id_data(331, 328))
    nvme0.sanitize().waitdone()  #L13

    with pytest.warns(UserWarning, match="ERROR status: 00/1d"):
        nvme0n1.write(qpair, buf, 0).waitdone()

    # check sanitize status in log page
    with pytest.warns(UserWarning, match="AER notification is triggered"):
        nvme0.getlogpage(0x81, buf, 20).waitdone()  #L17
        while buf.data(3, 2) & 0x7 != 1:  #L18
            time.sleep(1)
            nvme0.getlogpage(0x81, buf, 20).waitdone()  #L20
            progress = buf.data(1, 0)*100//0xffff
            logging.info("%d%%" % progress)

    # check sanitize status
    nvme0.getlogpage(0x81, buf, 20).waitdone()
    assert buf.data(3, 2) & 0x7 == 1





#
#  BSD LICENSE
#
#  Copyright (c) Crane Chu <cranechu@gmail.com>
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#    * Neither the name of Intel Corporation nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# -*- coding: utf-8 -*-


# for pypi package information
import setuptools
import subprocess
from setuptools.command.install import install

class CustomInstall(install):
    def run(self):
        subprocess.call("sudo dnf install -y make", shell=True)
        install.run(self)

with open("README.md", "r") as fh:
    long_description = fh.read()

    
setuptools.setup(
    name="pynvme",
    version="2.3.0",
    author="Crane Chu",
    author_email="cranechu@gmail.com",
    description="builds your own tests.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pynvme/pynvme",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: C",
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires='>=3.5',
    install_requires=['pytest',
                      'pytest-excel',
                      'libpci',
                      'pylspci',
                      'quarchpy',
                      'pytemperature'],
    data_files=[
        ('pynvme',
         ['nvme.so',
          'Makefile',
          'conftest.py',
          'driver_test.py',
          'pytest.ini']),
        ('pynvme/src',
         ['src/common.sh',
          'src/setup.sh']),
        ('pynvme/scripts',
         ['scripts/psd.py',
          'scripts/tcg.py',
          'scripts/test_examples.py',
          'scripts/test_utilities.py']),
        ('pynvme/scripts/stress',
         ['scripts/stress/dirty_power_cycle_test.py']),
        ('pynvme/scripts/performance',
         ['scripts/performance/1_fresh_perf_test.py',
          'scripts/performance/2_init_time_test.py',
          'scripts/performance/3_steady_perf_test.py',
          'scripts/performance/4_nvme_cmd_test.py',
          'scripts/performance/5_generate_report_test.py',
          'scripts/performance/report_template.xlsx'
          ]),
        ('pynvme/include/spdk',
         ['include/spdk/pci_ids.h']),
        ('pynvme/scripts/conformance/01_admin',
         ['scripts/conformance/01_admin/abort_test.py',
          'scripts/conformance/01_admin/identify_test.py',
          'scripts/conformance/01_admin/queue_test.py',
          'scripts/conformance/01_admin/firmware_test.py',
          'scripts/conformance/01_admin/dst_test.py',
          'scripts/conformance/01_admin/format_test.py',
          'scripts/conformance/01_admin/aer_test.py',
          'scripts/conformance/01_admin/sanitize_test.py',
          'scripts/conformance/01_admin/mi_test.py',
          'scripts/conformance/01_admin/features_test.py',
          'scripts/conformance/01_admin/logpage_test.py'
          ]),
        ('pynvme/scripts/conformance/02_nvm',
         ['scripts/conformance/02_nvm/compare_test.py',
          'scripts/conformance/02_nvm/flush_test.py',
          'scripts/conformance/02_nvm/read_test.py',
          'scripts/conformance/02_nvm/write_uncorrectable_test.py',
          'scripts/conformance/02_nvm/deallocate_test.py',
          'scripts/conformance/02_nvm/write_test.py',
          'scripts/conformance/02_nvm/verify_test.py',
          'scripts/conformance/02_nvm/write_zeroes_test.py'
          ]),
        ('pynvme/scripts/conformance/03_features',
         ['scripts/conformance/03_features/hmb_test.py',
          'scripts/conformance/03_features/write_protect_test.py',
          'scripts/conformance/03_features/power_management_test.py',
          'scripts/conformance/03_features/reset_test.py'
          ]),
        ('pynvme/scripts/conformance/04_registers',
         ['scripts/conformance/04_registers/controller_test.py',
          'scripts/conformance/04_registers/power_test.py',
          'scripts/conformance/04_registers/pcie_test.py'
          ]), 
        ('pynvme/scripts/conformance/05_controller',
         ['scripts/conformance/05_controller/sq_cq_test.py',
          'scripts/conformance/05_controller/sqe_cqe_test.py',
          'scripts/conformance/05_controller/interrupt_test.py',
          'scripts/conformance/05_controller/prp_test.py',
          'scripts/conformance/05_controller/arbitration_test.py'
          ]),
        ('pynvme/scripts/conformance/06_tcg',
         ['scripts/conformance/06_tcg/use_case_test.py'
          ]),
        ],
    cmdclass={'install': CustomInstall},
)



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging
import nvme as d

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


def test_create_cq_with_invalid_prp_offset(nvme0):
    prp = PRP(4096)
    #Spec NVM-Express-1_4-2019.06.10-Ratified
    #Figure 149: In both cases the PRP Entry shall have an offset of 0h.
    prp.offset = 0
    IOCQ(nvme0, 1, 10, prp).delete()

    #ERROR status: 00/13:invalid prp offset
    prp.offset = 2048
    with pytest.warns(UserWarning, match="ERROR status: 00/13"):
        IOCQ(nvme0, 1, 10, prp).delete()

    prp.offset = 2050
    with pytest.warns(UserWarning, match="ERROR status: 00/13"):
        IOCQ(nvme0, 1, 10, prp).delete()

    prp.offset = 4095
    with pytest.warns(UserWarning, match="ERROR status: 00/13"):
        IOCQ(nvme0, 2, 10, prp).delete()

    prp.offset = 255
    with pytest.warns(UserWarning, match="ERROR status: 00/13"):
        IOCQ(nvme0, 2, 10, prp).delete()


def test_create_sq_with_invalid_prp_offset(nvme0):
    prp = PRP(4096)
    cq = IOCQ(nvme0, 1, 10, prp)

    #Spec NVM-Express-1_4-2019.06.10-Ratified
    #Figure 153: In both cases, the PRP Entry shall have an offset of 0h.
    prp.offset = 0
    IOSQ(nvme0, 1, 10, prp, cqid=1).delete()

    prp.offset = 2048
    with pytest.warns(UserWarning, match="ERROR status: 00/13"):
        IOSQ(nvme0, 1, 10, prp, cqid=1).delete()

    prp.offset = 2050
    with pytest.warns(UserWarning, match="ERROR status: 00/13"):
        IOSQ(nvme0, 1, 10, prp, cqid=1).delete()

    prp.offset = 4095
    with pytest.warns(UserWarning, match="ERROR status: 00/13"):
        IOSQ(nvme0, 2, 10, prp, cqid=1).delete()

    prp.offset = 255
    with pytest.warns(UserWarning, match="ERROR status: 00/13"):
        IOSQ(nvme0, 2, 10, prp, cqid=1).delete()

    cq.delete()


@pytest.mark.parametrize("repeat", range(2))
def test_hello_world(nvme0, nvme0n1, repeat):
    # prepare data buffer and IO queue
    read_buf = Buffer(512)
    write_buf = Buffer(512)
    write_buf[10:21] = b'hello world'
    qpair = Qpair(nvme0, 16)  # create IO SQ/CQ pair, with 16 queue-depth

    # send write and read command
    def write_cb(cdw0, status1):  # command callback function
        nvme0n1.read(qpair, read_buf, 0, 1)
        assert status1 == 1  # phase-bit
    nvme0n1.write(qpair, write_buf, 0, 1, cb=write_cb)

    # wait commands complete and verify data
    assert read_buf[10:21] != b'hello world'
    qpair.waitdone(2)
    assert read_buf[10:21] == b'hello world'


def test_format_512(nvme0n1):
    nvme0n1.format(512)


@pytest.mark.parametrize("mdts", [64, 128, 256, 512, 800, 1024, 16*1024, 32*1024, 32*1024+64, 32*1024+64+8,64*1024])
def test_write_mdts(nvme0, mdts):
    cq = IOCQ(nvme0, 1, 2, PRP())
    sq = IOSQ(nvme0, 1, 2, PRP(), cqid=1)

    # prp for the long buffer
    write_buf_1 = PRP(ptype=32, pvalue=0xaaaaaaaa)
    pages = mdts//8
    pages -= 1

    prp_list = PRPList()
    prp_list_head = prp_list
    while pages:
        for i in range(511):
            if pages:
                prp_list[i] = PRP()
                pages -= 1
                logging.debug(pages)
        if pages>1:
            tmp = PRPList()
            prp_list[511] = tmp
            prp_list = tmp
            logging.debug("prp_list")
        elif pages==1:
            prp_list[511] = PRP()
            pages -= 1
            logging.debug(pages)

    w1 = SQE(1, 1)
    w1.prp1 = write_buf_1
    w1.prp2 = prp_list_head
    w1[12] = mdts-1 # 0based, nlba
    w1.cid = 0x123
    sq[0] = w1
    sq.tail = 1

    time.sleep(1)
    cqe = CQE(cq[0])
    logging.info(cqe)
    assert cqe.p == 1
    assert cqe.cid == 0x123
    assert cqe.sqhd == 1
    logging.info("cqe status is {}".format(cqe.status))
    assert cqe.status == 0 or cqe.status == 2
    cq.head = 1

    sq.delete()
    cq.delete()


@pytest.mark.parametrize("offset", [0, 4, 16, 32, 512, 800, 1024, 3000])
def test_page_offset(nvme0, nvme0n1, qpair, buf, offset):
    # fill the data
    write_buf = Buffer(512)
    nvme0n1.write(qpair, write_buf, 0x5aa5).waitdone()

    # read the data to different offset and check lba
    buf.offset = offset
    buf.size = 512
    nvme0n1.read(qpair, buf, 0x5aa5).waitdone()
    assert buf[offset] == 0xa5
    assert buf[offset+1] == 0x5a


@pytest.mark.parametrize("offset", [1, 2, 3, 501, 502])
def test_page_offset_invalid(nvme0, nvme0n1, qpair, offset):
    # fill the data
    write_buf = Buffer(512)
    nvme0n1.write(qpair, write_buf, 0xa5).waitdone()

    # read the data to different offset and check lba
    buf = Buffer(1024, ptype=0, pvalue=1)
    buf.offset = offset
    buf.size = 512

    # Spec NVM-Express-1_4-2019.06.10-Ratified
    # Figure 108:Note: The controller is not required to check that bits 1:0
    # are cleared to 00b. The controller may report an error of PRP Offset Invalid
    # pytest warning may not appear here
    nvme0n1.read(qpair, buf, 0xa5).waitdone()
    assert buf[offset] != 0xa5


@pytest.mark.parametrize("offset", [4, 16, 32, 512, 800, 1024, 3000])
def test_admin_page_offset(nvme0, offset):
    buf = d.Buffer(4096*2, 'controller identify data')
    buf.offset = offset
    buf.size = 4096
    assert buf[0] == 0
    assert buf[offset] == 0
    nvme0.identify(buf).waitdone()
    assert buf[offset] != 0
    assert buf[0] == 0


@pytest.mark.parametrize("offset", [1, 2, 3])
def test_admin_page_offset_invalid(nvme0, nvme0n1, qpair, offset):
    buf = d.Buffer(4096*2, 'controller identify data')
    buf.offset = offset
    buf.size = 32  # PRP1 only
    assert buf[0] == 0
    assert buf[offset] == 0
    # pytest warning may not appear here
    nvme0.identify(buf).waitdone()
    logging.info(buf.dump(16))
    assert buf[0] != 0


def test_valid_offset_prp_in_list(nvme0):
    cq = IOCQ(nvme0, 1, 10, PRP())
    sq = IOSQ(nvme0, 1, 10, PRP(), cqid=1)

    buf = PRP(ptype=32, pvalue=0xffffffff)
    buf.offset = 0x10

    prp_list = PRPList()
    prp_list.offset = 0x20

    for i in range(8):
        tmp = PRP(ptype=32, pvalue=0xffffffff)
        prp_list[i] = tmp

    logging.info(prp_list.dump(64))
    print(buf.dump(32))
    for i in range(8):
        print(prp_list[i].dump(32))

    cmd = SQE(2, 1)
    cmd.prp1 = buf
    cmd.prp2 = prp_list
    cmd[12] = 0x4000001f
    sq[0] = cmd
    logging.info(sq[0])
    sq.tail = 1
    while CQE(cq[0]).p == 0: pass
    cq.head = 1
    logging.info(cq[0])
    logging.info(hex(cq[0][3]>>17))
    print(buf.dump(32))
    for i in range(8):
        print(prp_list[i].dump(32))
    assert (cq[0][3]>>17)&0x3ff == 0x0

    sq.delete()
    cq.delete()


def test_invalid_offset_prp_in_list(nvme0):
    cq = IOCQ(nvme0, 1, 10, PRP())
    sq = IOSQ(nvme0, 1, 10, PRP(), cqid=1)

    buf = PRP(ptype=32, pvalue=0xffffffff)
    buf.offset = 0x10

    prp_list = PRPList()
    prp_list.offset = 0x20

    for i in range(8):
        tmp = PRP(ptype=32, pvalue=0xffffffff)
        tmp.offset = 0x10
        prp_list[i] = tmp

    logging.info(prp_list.dump(64))
    print(buf.dump(32))
    for i in range(8):
        print(prp_list[i].dump(32))

    cmd = SQE(2, 1)
    cmd.prp1 = buf
    cmd.prp2 = prp_list
    cmd[12] = 0x4000001f
    sq[0] = cmd
    logging.info(sq[0])
    sq.tail = 1
    while CQE(cq[0]).p == 0: pass
    cq.head = 1
    logging.info(cq[0])
    logging.info(hex(cq[0][3]>>17))
    print(buf.dump(32))
    for i in range(8):
        print(prp_list[i].dump(32))
    status = (cq[0][3]>>17)&0x3ff
    assert status == 0x0013 or status == 0

    sq.delete()
    cq.delete()





#
#  BSD LICENSE
#
#  Copyright (c) Crane Chu <cranechu@gmail.com>
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#    * Neither the name of Intel Corporation nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# -*- coding: utf-8 -*-


import time
import pytest
import nvme as d

import logging


rand = True
seq = False

read = True
write = False

def do_ioworker(rand, read, ns):
    """ run ioworkers for basic io performance tests"""
    
    seconds = 10
    io_size = 8 if rand else 128 # 4K or 64K
    rp = 100 if read else 0 # read or write

    r = ns.ioworker(io_size=io_size, lba_align=io_size,
                    region_end=(1<<30)//512, # 1GB space
                    lba_random=rand, qdepth=64,
                    read_percentage=rp, time=seconds).start().close()

    io_total = r.io_count_read+r.io_count_nonread
    iops = io_total//seconds

    return iops if rand else iops*io_size*512  # return Bps for seq IO


def do_fill_drive(rand, nvme0n1, nvme0, filename):
    io_size = 8 if rand else 128
    ns_size = nvme0n1.id_data(7, 0)
    io_count = ns_size//io_size
    io_per_second = []
    output_percentile_latency = dict.fromkeys([99.9])
    
    w = nvme0n1.ioworker(io_size=io_size, lba_align=io_size,
                         lba_random=rand, qdepth=64,
                         io_count=io_count, read_percentage=0,
                         output_percentile_latency=output_percentile_latency,
                         output_io_per_second=io_per_second).start()
    while w.running:
        # fixed the bug: aer in ioworker corrupted the aer counter in waitdone
        #nvme0.getfeatures(7).waitdone()
        time.sleep(1)
    r = w.close()
    
    import matplotlib.pyplot as plt
    plt.figure(figsize=(30, 12))
    plt.plot(r.latency_distribution)
    plt.xlabel('useconds')
    plt.ylabel('#IO')
    plt.xlim(1, len(r.latency_distribution))
    plt.ylim(bottom=1)
    plt.xscale('log')
    plt.yscale('log')
    plt.tight_layout()
    plt.savefig("lat_hist_"+filename+".png", dpi=300)
    plt.close()

    if not rand:
        return [iops*io_size*512 for iops in io_per_second]
    else:
        return io_per_second


def test_create_report_file(nvme0, nvme0n1, pcie):
    orig_timeout = nvme0.timeout
    nvme0.timeout = 100000
    
    nvme0n1.format(512)  # 512 sector size
    
    nvme0.timeout = orig_timeout

    import libpci
    vid = pcie.register(0, 2)
    vendor = libpci.LibPCI().lookup_vendor_name(vid)
    
    model = nvme0.id_data(63, 24, str)
    fw = nvme0.id_data(71, 64, str)
    capacity = str(nvme0n1.id_data(7, 0)*512)
    hmb = nvme0.id_data(275, 272) * 4
    
    qpairs = 0
    def getfeatures_cb(cdw0, status):
        nonlocal qpairs; qpairs = cdw0
        qpairs = (cdw0&0xffff)+1
    nvme0.getfeatures(7, cb=getfeatures_cb).waitdone()
    
    with open("report.csv", "w+") as f:
        f.write("%s\n" % vendor)
        f.write("%s\n" % model)
        f.write("%s\n" % fw)
        f.write("%s\n" % capacity)
        f.write("%d\n" % hmb)
        f.write("%d\n" % qpairs)
    

# empty read
def test_empty_read_performance(nvme0n1):
    logging.info(do_ioworker(seq, read, nvme0n1))
    logging.info(do_ioworker(rand, read, nvme0n1))

    
# write/read in 1GB, cdm
def test_1gb_read_write_performance(nvme0n1):
    with open("report.csv", "a") as f:
        f.write("%d\n" % do_ioworker(seq, write, nvme0n1))
        f.write("%d\n" % do_ioworker(seq, read, nvme0n1))
        f.write("%d\n" % do_ioworker(rand, write, nvme0n1))
        f.write("%d\n" % do_ioworker(rand, read, nvme0n1))


# full drive seq write
def test_fill_drive_first_pass(nvme0, nvme0n1):
    io_per_sec = do_fill_drive(seq, nvme0n1, nvme0, "first_seq")
    io_per_sec = io_per_sec[:300]
    with open("report.csv", "a") as f:
        for iops in io_per_sec:
            f.write('%d\n' % iops)
        for i in range(300):
            f.write('0\n')
    
# random
def test_fill_drive_random(nvme0, nvme0n1):
    io_per_sec = do_fill_drive(rand, nvme0n1, nvme0, "first_rand")
    io_per_sec = io_per_sec[:600]
    with open("report.csv", "a") as f:
        for iops in io_per_sec:
            f.write('%d\n' % iops)
            
        # add temperature, °C
        import pytemperature
        logpage_buf = d.Buffer(512)
        nvme0.getlogpage(2, logpage_buf).waitdone()
        t = round(pytemperature.k2c(logpage_buf.data(2, 1)))
        logging.info(t)
        f.write('%d\n' % t)

        
# 2-pass full drive seq write
@pytest.mark.parametrize("repeat", range(2))
def test_fill_drive_after_random(nvme0, nvme0n1, repeat):
    io_per_sec = do_fill_drive(seq, nvme0n1, nvme0, "seq_after_random_%d"%repeat)
    io_per_sec = io_per_sec[:600]
    with open("report.csv", "a") as f:
        for iops in io_per_sec:
            f.write('%d\n' % iops)
